import numpy as np
import pytest
import astropy.units as u
from astropy.nddata import NDData
from astropy.coordinates import SkyCoord
from regions import PixCoord, CirclePixelRegion, CircleSkyRegion, RectangleSkyRegion

from jdaviz.core.marks import FootprintOverlay
from jdaviz.configs.imviz.plugins.footprints.preset_regions import _all_apertures
from astropy.wcs import WCS

pytest.importorskip("pysiaf")


def _get_markers_from_viewer(viewer):
    if hasattr(viewer, '_obj'):
        viewer = viewer._obj
    return [m for m in viewer.figure.marks if isinstance(m, FootprintOverlay)]


def test_user_api(imviz_helper, image_2d_wcs, tmp_path):
    arr = np.ones((10, 10))
    ndd = NDData(arr, wcs=image_2d_wcs)
    # load the image twice to test linking
    imviz_helper.load_data(ndd)
    imviz_helper.load_data(ndd)

    plugin = imviz_helper.plugins['Footprints']
    default_color = plugin.color
    default_opacity = plugin.fill_opacity
    plugin.color = '#333333'
    with plugin.as_active():
        assert plugin._obj.is_pixel_linked is True
        plugin._obj.vue_link_by_wcs()
        assert plugin._obj.is_pixel_linked is False

        # test that each of the supported instruments/presets work
        for preset in (preset for preset in plugin.preset.choices if preset != 'From File...'):
            plugin.preset = preset

            viewer_marks = _get_markers_from_viewer(imviz_helper.default_viewer)
            assert len(viewer_marks) == len(_all_apertures.get(preset))

        # regression test for user-set traitlets (specifically color) being reset
        # when the plugin is opened
        assert plugin.color == '#333333'
        assert viewer_marks[0].colors == ['#333333']

        # change the color/opacity of the first/default overlay
        plugin.color = '#ffffff'
        plugin.fill_opacity = 0.5

        viewer_marks = _get_markers_from_viewer(imviz_helper.default_viewer)
        assert viewer_marks[0].visible is True
        assert viewer_marks[0].colors == ['#ffffff']
        assert viewer_marks[0].fill_opacities == [0.5]

        # add a second overlay from the API
        with pytest.raises(ValueError):
            plugin.add_overlay(plugin.overlay.selected)

        plugin.add_overlay('second overlay')
        assert plugin.overlay.selected == 'second overlay'

        # color should reset, opacity should carry over
        assert plugin.color == default_color
        assert plugin.fill_opacity != default_opacity

        plugin.color = '#000000'

        # rename overlay
        plugin.rename_overlay('second overlay', 'renamed overlay')
        assert plugin.overlay.selected == 'renamed overlay'
        assert plugin.color == '#000000'

        # remove overlay, selection should default to first entry and traitlets update
        plugin.remove_overlay('renamed overlay')
        assert plugin.overlay.selected == 'default'
        assert plugin.color == '#ffffff'

        # test toggling visibility of markers
        viewer_marks = _get_markers_from_viewer(imviz_helper.default_viewer)
        assert viewer_marks[0].visible is True
        plugin.visible = False
        assert viewer_marks[0].visible is False
        plugin.visible = True
        assert viewer_marks[0].visible is True
        plugin.viewer.selected = []
        assert viewer_marks[0].visible is False
        plugin.viewer.select_all()
        assert viewer_marks[0].visible is True

        # test centering logic (for now just that it doesn't fail)
        plugin.center_on_viewer()

        # test from file/API ability
        reg = plugin.overlay_regions
        plugin.import_region(reg)
        assert plugin.preset.selected == 'From File...'
        viewer_marks = _get_markers_from_viewer(imviz_helper.default_viewer)
        assert len(viewer_marks) == len(reg)
        # test that importing a different region updates the marks and also that
        # a single region is supported
        plugin.import_region(reg[0])
        viewer_marks = _get_markers_from_viewer(imviz_helper.default_viewer)
        assert len(viewer_marks) == 1
        # clearing the file should default to the PREVIOUS preset (last from the for-loop above)
        plugin._obj.vue_file_import_cancel()
        assert plugin.preset.selected == preset

        tmp_file = str(tmp_path / 'test_region.reg')
        reg.write(tmp_file, format='ds9')
        plugin.import_region(tmp_file)
        assert plugin.preset.selected == 'From File...'
        assert plugin.preset.from_file == tmp_file

        # test single region (footprints contain multiple regions)
        valid_region_sky = CircleSkyRegion(center=SkyCoord(42, 43, unit='deg', frame='fk5'),
                                           radius=3 * u.deg)
        plugin.import_region(valid_region_sky)

        # test RectangleSkyRegion -> RectanglePixelRegion
        valid_region_sky = RectangleSkyRegion(center=SkyCoord(42, 43, unit='deg', frame='fk5'),
                                              height=3 * u.deg, width=2 * u.deg)
        plugin.import_region(valid_region_sky)

        tmp_invalid_path = str(tmp_path / 'invalid_path.reg')
        with pytest.raises(ValueError):
            plugin.import_region(tmp_invalid_path)
        with pytest.raises(TypeError):
            plugin.import_region(5)

        invalid_region = CirclePixelRegion(PixCoord(x=8, y=7), radius=3.5)
        with pytest.raises(ValueError):
            plugin.import_region(invalid_region)
        tmp_invalid_file = str(tmp_path / 'test_invalid_region.reg')
        invalid_region.write(tmp_invalid_file, format='ds9')
        with pytest.raises(ValueError):
            plugin.import_region(tmp_invalid_file)
        assert plugin.preset.selected == preset

    # with the plugin no longer active, marks should not be visible
    assert plugin._obj.is_active is False
    viewer_marks = _get_markers_from_viewer(imviz_helper.default_viewer)
    assert viewer_marks[0].visible is False


def test_api_after_linking(imviz_helper):
    # custom image to enable visual test in a notebook
    arr = np.ones((300, 300))
    image_2d_wcs = WCS({'CTYPE1': 'RA---TAN', 'CUNIT1': 'deg', 'CDELT1': -0.0002777777778,
                        'CRPIX1': 1, 'CRVAL1': 337.5202808,
                        'CTYPE2': 'DEC--TAN', 'CUNIT2': 'deg', 'CDELT2': 0.0002777777778,
                        'CRPIX2': 1, 'CRVAL2': -20.833333059999998})

    viewer = imviz_helper.app.get_viewer_by_id('imviz-0')

    ndd = NDData(arr, wcs=image_2d_wcs)
    imviz_helper.load_data(ndd)
    imviz_helper.load_data(ndd)

    plugin = imviz_helper.plugins['Footprints']
    with plugin.as_active():

        pointing = {'name': ['pt'], 'ra': [337.51], 'dec': [-20.77], 'pa': [0]}
        plugin.color = 'green'
        plugin.add_overlay(pointing['name'][0])
        plugin.ra = pointing['ra'][0]
        plugin.dec = pointing['dec'][0]
        plugin.pa = pointing['pa'][0]

        # when pixel linked are any marks displayed
        viewer_marks = _get_markers_from_viewer(viewer)
        no_marks_displayed = all(not mark.visible for mark in viewer_marks)
        assert no_marks_displayed is True

        # link by wcs and retest
        imviz_helper.link_data(align_by='wcs')

        viewer_marks = _get_markers_from_viewer(viewer)
        # distinguish default from custom overlay with color
        marks_displayed = any(mark.visible and mark.colors[0] == 'green'
                              for mark in viewer_marks)

        # when wcs linked are any marks displayed
        assert marks_displayed is True


def test_footprint_updates_on_rotation(imviz_helper):
    image_2d_wcs = WCS({'CTYPE1': 'RA---TAN', 'CUNIT1': 'deg',
                        'CRPIX1': 1, 'CRVAL1': 337.5202808,
                        'CTYPE2': 'DEC--TAN', 'CUNIT2': 'deg',
                        'CRPIX2': 1, 'CRVAL2': -20.833333059999998,
                        'CD1_1': 0.0005, 'CD1_2': 0.005,
                        'CD2_1': -0.005, 'CD2_2': -0.0005})

    arr = np.random.normal(size=(10, 10))
    ndd = NDData(arr, wcs=image_2d_wcs)

    imviz_helper.load_data(ndd)
    imviz_helper.link_data(align_by='wcs')

    footprints = imviz_helper.plugins['Footprints']
    footprints.keep_active = True

    # this is near the pixel origin:
    rectangle_center = SkyCoord(337.5202808, -20.83333306, unit='deg')

    # this is the opposite corner of the image:
    opposite_corner = SkyCoord(337.5791498, -20.88832297, unit='deg')

    # Put a MIRI footprint at the far corner:
    footprints.preset = 'MIRI'
    footprints.ra = opposite_corner.ra.deg
    footprints.dec = opposite_corner.dec.deg
    miri_region = footprints.overlay_regions[0]

    # create a second footprint for a rectangular sky region near the pixel origin:
    footprints.add_overlay('2')

    rectangle_region = RectangleSkyRegion(
        center=rectangle_center,
        height=1 * u.arcmin,
        width=2 * u.arcmin,
        angle=45 * u.deg
    )

    footprints.import_region(rectangle_region)

    # ensure that the footprints are where we expect them:
    assert rectangle_region.contains(rectangle_center, image_2d_wcs)
    assert not rectangle_region.contains(opposite_corner, image_2d_wcs)

    assert not miri_region.contains(rectangle_center, image_2d_wcs)
    assert miri_region.contains(opposite_corner, image_2d_wcs)

    marks = _get_markers_from_viewer(imviz_helper.default_viewer)

    # check that the rectangle region appears near the bottom of the viewer:
    assert np.concatenate([marks[0].y, marks[1].y]).min() < -3

    # now rotate to north-up east-left:
    orientation = imviz_helper.plugins['Orientation']
    orientation.set_north_up_east_left()

    # If all footprint orientations have been updated, the lowest
    # mark should still be centered low. If the footprint
    # orientations aren't updated, both footprints will be
    # at the top of the viewer, and this test will fail.
    marks = _get_markers_from_viewer(imviz_helper.default_viewer)
    assert np.concatenate([marks[0].y, marks[1].y]).min() < -3



from functools import cached_property
from regions import Regions
from specutils import SpectralRegion

from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry


__all__ = ['RegionsParser']


@loader_parser_registry('Regions')
class RegionsParser(BaseParser):
    @property
    def is_valid(self):
        if self.app.config != 'specviz':
            # NOTE: temporary during deconfig process
            return False
        if isinstance(self.input, str):
            ext = self.input.split('.')[-1]
            # once deconfigging is complete:
            # return ext in ('reg', 'fits', 'ecsv')
            return ext == 'ecsv'

    @cached_property
    def output(self):
        region_format = None
        try:
            return Regions.read(self.input, format=region_format)
        except Exception:  # nosec
            return SpectralRegion.read(self.input)



import os
import warnings

import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.time import Time
from astropy.wcs import WCS
from specutils import Spectrum1D

from jdaviz.core.custom_units_and_equivs import PIX2, _eqv_flux_to_sb_pixel
from jdaviz.core.registries import data_parser_registry
from jdaviz.core.unit_conversion_utils import check_if_unit_is_per_solid_angle
from jdaviz.utils import standardize_metadata, PRIHDR_KEY, download_uri_to_path

__all__ = ['parse_data']

EXT_TYPES = dict(flux=['flux', 'sci', 'data'],
                 uncert=['ivar', 'err', 'error', 'var', 'uncert'],
                 mask=['mask', 'dq', 'quality', 'data_quality'])


@data_parser_registry("cubeviz-data-parser")
def parse_data(app, file_obj, data_type=None, data_label=None,
               parent=None, cache=None, local_path=None, timeout=None):
    """
    Attempts to parse a data file and auto-populate available viewers in
    cubeviz.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    file_obj : str
        The path to a cube-like data file.
    data_type : str, {'flux', 'mask', 'uncert'}
        The data type used to explicitly differentiate parsed data.
    data_label : str, optional
        The label to be applied to the Glue data component.
    parent : str, optional
        Data label for "parent" data to associate with the loaded data as "child".
    cache : None, bool, or str
        Cache the downloaded file if the data are retrieved by a query
        to a URL or URI.
    local_path : str, optional
        Cache remote files to this path. This is only used if data is
        requested from `astroquery.mast`.
    timeout : float, optional
        If downloading from a remote URI, set the timeout limit for
        remote requests in seconds (passed to
        `~astropy.utils.data.download_file` or
        `~astroquery.mast.Conf.timeout`).
    """

    flux_viewer_reference_name = app._jdaviz_helper._default_flux_viewer_reference_name
    uncert_viewer_reference_name = app._jdaviz_helper._default_uncert_viewer_reference_name
    spectrum_viewer_reference_name = app._jdaviz_helper._default_spectrum_viewer_reference_name

    if data_type is not None and data_type.lower() not in ('flux', 'mask', 'uncert'):
        raise TypeError("Data type must be one of 'flux', 'mask', or 'uncert' "
                        f"but got '{data_type}'")

    # If the file object is an hdulist or a string, use the generic parser for
    #  fits files.
    # TODO: this currently only supports fits files. We will want to make this
    #  generic enough to work with other file types (e.g. ASDF). For now, this
    #  supports MaNGA and JWST data.
    if isinstance(file_obj, fits.hdu.hdulist.HDUList):
        try:
            _parse_spectrum1d_3d(
                app, Spectrum1D.read(file_obj), data_label=data_label,
                flux_viewer_reference_name=flux_viewer_reference_name,
                uncert_viewer_reference_name=uncert_viewer_reference_name
            )
        except Exception:  # nosec
            _parse_hdulist(
                app, file_obj, file_name=data_label,
                flux_viewer_reference_name=flux_viewer_reference_name,
                uncert_viewer_reference_name=uncert_viewer_reference_name
            )
    elif isinstance(file_obj, str):
        if file_obj.lower().endswith('.gif'):  # pragma: no cover
            _parse_gif(app, file_obj, data_label,
                       flux_viewer_reference_name=flux_viewer_reference_name)
            return

        # try parsing file_obj as a URI/URL:
        file_obj = download_uri_to_path(
            file_obj, cache=cache, local_path=local_path, timeout=timeout
        )

        file_name = os.path.basename(file_obj)

        with fits.open(file_obj) as hdulist:
            prihdr = hdulist[0].header
            telescop = prihdr.get('TELESCOP', '').lower()
            exptype = prihdr.get('EXP_TYPE', '').lower()
            # NOTE: Alerted to deprecation of FILETYPE keyword from pipeline on 2022-07-08
            # Kept for posterity in for data processed prior to this date. Use EXP_TYPE instead
            filetype = prihdr.get('FILETYPE', '').lower()
            system = prihdr.get('SYSTEM', '').lower()
            if telescop == 'jwst' and ('ifu' in exptype or
                                       'mrs' in exptype or
                                       filetype == '3d ifu cube'):
                for ext, viewer_name in (('SCI', flux_viewer_reference_name),
                                         ('ERR', uncert_viewer_reference_name),
                                         ('DQ', None)):
                    data_label = app.return_data_label(file_name, ext)

                    if ext == 'SCI':
                        sci_ext = data_label

                    # TODO: generalize/centralize this for use in other configs too

                    if parent is not None:
                        parent_data_label = parent
                    elif ext == 'DQ':
                        parent_data_label = sci_ext
                    else:
                        parent_data_label = None

                    _parse_jwst_s3d(
                        app, hdulist, data_label, ext=ext, viewer_name=viewer_name,
                        flux_viewer_reference_name=flux_viewer_reference_name,
                        parent=parent_data_label
                    )
            elif telescop == 'jwst' and filetype == 'r3d' and system == 'esa-pipeline':
                for ext, viewer_name in (('DATA', flux_viewer_reference_name),
                                         ('ERR', uncert_viewer_reference_name),
                                         ('QUALITY', None)):
                    data_label = app.return_data_label(file_name, ext)
                    _parse_esa_s3d(
                        app, hdulist, data_label, ext=ext, viewer_name=viewer_name,
                        flux_viewer_reference_name=flux_viewer_reference_name,
                    )
            else:
                try:
                    _parse_spectrum1d_3d(
                        app, Spectrum1D.read(hdulist), data_label=data_label or file_name,
                        flux_viewer_reference_name=flux_viewer_reference_name,
                        uncert_viewer_reference_name=uncert_viewer_reference_name
                    )
                except Exception:  # nosec
                    _parse_hdulist(
                        app, hdulist, file_name=data_label or file_name,
                        flux_viewer_reference_name=flux_viewer_reference_name,
                        uncert_viewer_reference_name=uncert_viewer_reference_name
                    )

    # If the data types are custom data objects, use explicit parsers. Note
    #  that this relies on the glue-astronomy machinery to turn the data object
    #  into something glue can understand.
    elif isinstance(file_obj, Spectrum1D) and file_obj.flux.ndim in (1, 3):
        if file_obj.flux.ndim == 3:
            _parse_spectrum1d_3d(
                app, file_obj, data_label=data_label,
                flux_viewer_reference_name=flux_viewer_reference_name,
                uncert_viewer_reference_name=uncert_viewer_reference_name
            )
        else:
            _parse_spectrum1d(
                app, file_obj, data_label=data_label,
                spectrum_viewer_reference_name=spectrum_viewer_reference_name
            )

    elif isinstance(file_obj, np.ndarray) and file_obj.ndim == 3:
        _parse_ndarray(app, file_obj, data_label=data_label, data_type=data_type,
                       flux_viewer_reference_name=flux_viewer_reference_name,
                       uncert_viewer_reference_name=uncert_viewer_reference_name)
    else:
        raise NotImplementedError(f'Unsupported data format: {file_obj}')


def _get_celestial_wcs(wcs):
    """ If `wcs` has a celestial component return that, otherwise return None """
    return wcs.celestial if hasattr(wcs, 'celestial') else None


def _return_spectrum_with_correct_units(flux, wcs, metadata, data_type=None,
                                        target_wave_unit=None, hdulist=None,
                                        uncertainty=None, mask=None, apply_pix2=False):
    """Upstream issue of WCS not using the correct units for data must be fixed here.
    Issue: https://github.com/astropy/astropy/issues/3658.

    Also converts flux units to flux/pix2 solid angle units, if `flux` is not a surface
    brightness and `apply_pix2` is True.
    """
    # handle scale factors when they are included in the unit
    # (has to be done before Spectrum1D creation)
    if not np.isclose(flux.unit.scale, 1, rtol=1e-5):
        flux = flux.to(flux.unit / flux.unit.scale)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', message='Input WCS indicates that the spectral axis is not last',
            category=UserWarning)
        sc = Spectrum1D(flux=flux, wcs=wcs, meta=metadata, uncertainty=uncertainty, mask=mask)

    # convert flux and uncertainty to per-pix2 if input is not a surface brightness
    target_flux_unit = None
    if (apply_pix2 and (data_type != "mask") and
            (not check_if_unit_is_per_solid_angle(flux.unit))):
        target_flux_unit = flux.unit / PIX2
    elif check_if_unit_is_per_solid_angle(flux.unit, return_unit=True) == "spaxel":
        # We need to convert spaxel to pixel squared, since spaxel isn't fully supported by astropy
        # This is horribly ugly but just multiplying by u.Unit("spaxel") doesn't work
        target_flux_unit = flux.unit * u.Unit('spaxel') / PIX2

    if target_wave_unit is None and hdulist is not None:
        found_target = False
        for ext in ('SCI', 'FLUX', 'PRIMARY', 'DATA'):  # In priority order
            if found_target:
                break
            if ext not in hdulist:
                continue
            hdr = hdulist[ext].header
            # The WCS could be swapped or unswapped.
            for cunit_num in (3, 1):
                cunit_key = f"CUNIT{cunit_num}"
                ctype_key = f"CTYPE{cunit_num}"
                if cunit_key in hdr and 'WAV' in hdr[ctype_key]:
                    target_wave_unit = u.Unit(hdr[cunit_key])
                    found_target = True
                    break

    if target_wave_unit == sc.spectral_axis.unit:
        target_wave_unit = None

    if (target_wave_unit is None) and (target_flux_unit is None):  # Nothing to convert
        new_sc = sc
    elif target_flux_unit is None:  # Convert wavelength only
        new_sc = sc.with_spectral_axis_unit(target_wave_unit)
    elif target_wave_unit is None:  # Convert flux only and only PIX2 stuff
        new_sc = sc.with_flux_unit(target_flux_unit, equivalencies=_eqv_flux_to_sb_pixel())
    else:  # Convert both
        new_sc = sc.with_spectral_axis_and_flux_units(
            target_wave_unit, target_flux_unit, flux_equivalencies=_eqv_flux_to_sb_pixel())

    if target_wave_unit is not None:
        new_sc.meta['_orig_spec'] = sc  # Need this for later

    return new_sc


def _parse_hdulist(app, hdulist, file_name=None,
                   flux_viewer_reference_name=None,
                   uncert_viewer_reference_name=None):
    if file_name is None and hasattr(hdulist, 'file_name'):
        file_name = hdulist.file_name
    else:
        file_name = file_name or "Unknown HDU object"

    is_loaded = []
    wcs_sci = None

    # TODO: This needs refactoring to be more robust.
    # Current logic fails if there are multiple EXTVER.
    for hdu in hdulist:
        if hdu.data is None or not hdu.is_image or hdu.data.ndim != 3:
            continue

        data_type = _get_data_type_by_hdu(hdu)
        if not data_type:
            continue

        # Only load each type once.
        if data_type in is_loaded:
            continue

        is_loaded.append(data_type)
        data_label = app.return_data_label(file_name, hdu.name)

        if data_type == 'flux':
            wcs = WCS(hdu.header, hdulist)
            wcs_sci = wcs
        else:
            wcs = wcs_sci

        if 'BUNIT' in hdu.header:
            try:
                flux_unit = u.Unit(hdu.header['BUNIT'])
            except Exception:
                warnings.warn("Invalid BUNIT, using count as data unit", UserWarning)
                flux_unit = u.count
        elif data_type == 'mask':  # DQ flags have no unit
            flux_unit = u.dimensionless_unscaled
        else:
            warnings.warn("Invalid BUNIT, using count as data unit", UserWarning)
            flux_unit = u.count

        flux = hdu.data << flux_unit

        metadata = standardize_metadata(hdu.header)
        if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
            metadata[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)

        # store original WCS in metadata. this is a hacky workaround for converting subsets
        # to sky regions, where the parent data of the subset might have dropped spatial WCS info
        metadata['_orig_spatial_wcs'] = _get_celestial_wcs(wcs)

        apply_pix2 = data_type in ['flux', 'uncert']
        sc = _return_spectrum_with_correct_units(flux, wcs, metadata, data_type=data_type,
                                                 hdulist=hdulist, apply_pix2=apply_pix2)

        app.add_data(sc, data_label)

        if data_type == 'mask':
            # We no longer auto-populate the mask cube into a viewer, but we still want
            # to keep track of this cube for use in, e.g., spectral extraction.
            app._jdaviz_helper._loaded_mask_cube = app.data_collection[data_label]

        elif data_type == 'uncert':
            app.add_data_to_viewer(uncert_viewer_reference_name, data_label)
            app._jdaviz_helper._loaded_uncert_cube = app.data_collection[data_label]

        else:  # flux
            # Forced wave unit conversion made it lose stuff, so re-add
            # also re-get unit from sc in case a factor of pix2 was applied
            app.data_collection[data_label].get_component("flux").units = sc.unit
            # Add flux to top left image viewer
            app.add_data_to_viewer(flux_viewer_reference_name, data_label)
            app._jdaviz_helper._loaded_flux_cube = app.data_collection[data_label]


def _parse_jwst_s3d(app, hdulist, data_label, ext='SCI',
                    viewer_name=None, flux_viewer_reference_name=None,
                    parent=None):
    hdu = hdulist[ext]
    data_type = _get_data_type_by_hdu(hdu)

    # Manually inject MJD-OBS until we can support GWCS, see
    # https://github.com/spacetelescope/jdaviz/issues/690 and
    # https://github.com/glue-viz/glue-astronomy/issues/59
    if ext == 'SCI' and 'MJD-OBS' not in hdu.header:
        for key in ('MJD-BEG', 'DATE-OBS'):  # Possible alternatives
            if key in hdu.header:
                if key.startswith('MJD'):
                    hdu.header['MJD-OBS'] = hdu.header[key]
                    break
                else:
                    t = Time(hdu.header[key])
                    hdu.header['MJD-OBS'] = t.mjd
                    break

    if ext == 'DQ':  # DQ flags have no unit
        flux = hdu.data << u.dimensionless_unscaled
    else:
        unit = u.Unit(hdu.header.get('BUNIT', 'count'))
        flux = hdu.data << unit
    wcs = WCS(hdulist['SCI'].header, hdulist)  # Everything uses SCI WCS

    metadata = standardize_metadata(hdu.header)

    # store original WCS in metadata. this is a hacky workaround for converting subsets
    # to sky regions, where the parent data of the subset might have dropped spatial WCS info
    metadata['_orig_spatial_wcs'] = _get_celestial_wcs(wcs)

    if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
        metadata[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)

    data = _return_spectrum_with_correct_units(
        flux, wcs, metadata, data_type=data_type, hdulist=hdulist)
    app.add_data(data, data_label, parent=parent)

    # get glue data and update if DQ:
    if ext == 'DQ':
        # prevent circular import:
        from jdaviz.configs.imviz.plugins.parsers import prep_data_layer_as_dq

        data = app.data_collection[-1]
        prep_data_layer_as_dq(data)

    if data_type == 'flux':  # Forced wave unit conversion made it lose stuff, so re-add
        app.data_collection[-1].get_component("flux").units = flux.unit

    if viewer_name is not None:
        app.add_data_to_viewer(viewer_name, data_label)

    if ext == 'DQ':
        app.add_data_to_viewer(flux_viewer_reference_name, data_label, visible=False)

    if data_type == 'flux':
        app._jdaviz_helper._loaded_flux_cube = app.data_collection[data_label]
    elif data_type == 'uncert':
        app._jdaviz_helper._loaded_uncert_cube = app.data_collection[data_label]


def _parse_esa_s3d(app, hdulist, data_label, ext='DATA', flux_viewer_reference_name=None):
    hdu = hdulist[ext]
    data_type = _get_data_type_by_hdu(hdu)

    if ext == 'QUALITY':  # QUALITY flags have no unit
        flux = hdu.data << u.dimensionless_unscaled
    else:
        unit = u.Unit(hdu.header.get('BUNIT', 'count'))
        flux = hdu.data << unit

    hdr = hdulist[1].header

    wcs_dict = {
        'CTYPE1': 'WAVE    ', 'CUNIT1': 'um', 'CDELT1': hdr['CDELT3'] * 1E6,
        'CRPIX1': hdr['CRPIX3'],
        'CRVAL1': hdr['CRVAL3'] * 1E6, 'NAXIS1': hdr['NAXIS3'],
        'CTYPE2': 'DEC--TAN', 'CUNIT2': 'deg', 'CDELT2': hdr['CDELT1'], 'CRPIX2': hdr['CRPIX1'],
        'CRVAL2': hdr['CRVAL1'], 'NAXIS2': hdr['NAXIS1'],
        'CTYPE3': 'RA---TAN', 'CUNIT3': 'deg', 'CDELT3': hdr['CDELT2'], 'CRPIX3': hdr['CRPIX2'],
        'CRVAL3': hdr['CRVAL2'], 'NAXIS3': hdr['NAXIS2']}

    wcs = WCS(wcs_dict)
    flux = np.moveaxis(flux, 0, -1)
    flux = np.swapaxes(flux, 0, 1)

    metadata = standardize_metadata(hdu.header)
    metadata.update(wcs_dict)  # To be internally consistent
    if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
        metadata[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)

    # store original WCS in metadata. this is a hacky workaround for converting subsets
    # to sky regions, where the parent data of the subset might have dropped spatial WCS info
    metadata['_orig_spatial_wcs'] = _get_celestial_wcs(wcs)

    data = _return_spectrum_with_correct_units(
        flux, wcs, metadata, data_type=data_type, hdulist=hdulist)

    app.add_data(data, data_label)

    if data_type == 'flux':  # Forced wave unit conversion made it lose stuff, so re-add
        app.data_collection[-1].get_component("flux").units = flux.unit

    app.add_data_to_viewer(flux_viewer_reference_name, data_label)

    if data_type == 'flux':
        app._jdaviz_helper._loaded_flux_cube = app.data_collection[data_label]
    elif data_type == 'uncert':
        app._jdaviz_helper._loaded_uncert_cube = app.data_collection[data_label]
    elif data_type == 'mask':
        app._jdaviz_helper._loaded_mask_cube = app.data_collection[data_label]


def _parse_spectrum1d_3d(app, file_obj, data_label=None,
                         flux_viewer_reference_name=None,
                         uncert_viewer_reference_name=None):
    """Load spectrum1d as a cube."""

    if data_label is None:
        data_label = "Unknown spectrum object"

    for attr in ("flux", "mask", "uncertainty"):
        val = getattr(file_obj, attr)
        if val is None:
            continue

        if attr == "mask":
            flux = val << u.dimensionless_unscaled  # DQ flags have no unit
        elif attr == "uncertainty":
            flux = val.represent_as(StdDevUncertainty).quantity
            flux[np.isinf(flux)] = np.nan  # Avoid INF from IVAR conversion
        else:
            flux = val

        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', message='Input WCS indicates that the spectral axis is not last',
                category=UserWarning)
            meta = standardize_metadata(file_obj.meta)

            # store original WCS in metadata. this is a hacky workaround for
            # converting subsets to sky regions, where the parent data of the
            # subset might have dropped spatial WCS info
            meta['_orig_spatial_wcs'] = None
            if hasattr(file_obj, 'wcs'):
                meta['_orig_spatial_wcs'] = _get_celestial_wcs(file_obj.wcs)

            # Also convert data loaded in flux units to a per-square-pixel surface
            # brightness unit (e.g Jy to Jy/pix**2)
            s1d = _return_spectrum_with_correct_units(
                flux, file_obj.wcs, meta, data_type=attr, apply_pix2=True)

        cur_data_label = app.return_data_label(data_label, attr.upper())
        app.add_data(s1d, cur_data_label)

        if attr == 'flux':
            app.add_data_to_viewer(flux_viewer_reference_name, cur_data_label)
            app._jdaviz_helper._loaded_flux_cube = app.data_collection[cur_data_label]
        elif attr == 'uncertainty':
            app.add_data_to_viewer(uncert_viewer_reference_name, cur_data_label)
            app._jdaviz_helper._loaded_uncert_cube = app.data_collection[cur_data_label]
        elif attr == 'mask':
            app._jdaviz_helper._loaded_mask_cube = app.data_collection[cur_data_label]


def _parse_spectrum1d(app, file_obj, data_label=None, spectrum_viewer_reference_name=None):

    # Here 'file_obj' is a Spectrum1D

    if data_label is None:
        data_label = app.return_data_label(file_obj)

    # store original WCS in metadata. this is a hacky workaround for converting subsets
    # to sky regions, where the parent data of the subset might have dropped spatial WCS info
    file_obj.meta['_orig_spatial_wcs'] = _get_celestial_wcs(file_obj.wcs) if hasattr(file_obj, 'wcs') else None  # noqa: E501

    # TODO: glue-astronomy translators only look at the flux property of
    #  specutils Spectrum1D objects. Fix to support uncertainties and masks.

    # convert data loaded in flux units to a per-square-pixel surface
    # brightness unit (e.g Jy to Jy/pix**2)
    if not check_if_unit_is_per_solid_angle(file_obj.flux.unit):
        file_obj = file_obj.with_flux_unit(
            file_obj.flux.unit / PIX2, equivalencies=_eqv_flux_to_sb_pixel())

    app.add_data(file_obj, data_label)
    app.add_data_to_viewer(spectrum_viewer_reference_name, data_label)


def _parse_ndarray(app, file_obj, data_label=None, data_type=None,
                   flux_viewer_reference_name=None,
                   uncert_viewer_reference_name=None):
    if data_label is None:
        data_label = app.return_data_label(file_obj)

    if data_type is None:
        data_type = 'flux'

    # Cannot change axis to ensure roundtripping within Cubeviz.
    # Axes must already be (x, y, z) at this point.
    flux = file_obj

    if not hasattr(flux, 'unit'):
        flux = flux << (u.count / PIX2)

    meta = standardize_metadata({'_orig_spatial_wcs': None})
    s3d = Spectrum1D(flux=flux, meta=meta)

    # convert data loaded in flux units to a per-square-pixel surface
    # brightness unit (e.g Jy to Jy/pix**2)
    if not check_if_unit_is_per_solid_angle(s3d.flux.unit):
        s3d = s3d.with_flux_unit(s3d.flux.unit / PIX2, equivalencies=_eqv_flux_to_sb_pixel())

    app.add_data(s3d, data_label)

    if data_type == 'flux':
        app.add_data_to_viewer(flux_viewer_reference_name, data_label)
        app._jdaviz_helper._loaded_flux_cube = app.data_collection[data_label]
    elif data_type == 'uncert':
        app.add_data_to_viewer(uncert_viewer_reference_name, data_label)
        app._jdaviz_helper._loaded_uncert_cube = app.data_collection[data_label]
    elif data_type == 'mask':
        app._jdaviz_helper._loaded_mask_cube = app.data_collection[data_label]


def _parse_gif(app, file_obj, data_label=None, flux_viewer_reference_name=None):  # pragma: no cover
    # NOTE: Parsing GIF needs imageio and Pillow, both are which undeclared
    # in setup.cfg but might or might not be installed by declared ones.
    import imageio

    file_name = os.path.basename(file_obj)

    if data_label is None:
        data_label = app.return_data_label(file_obj)

    flux = imageio.v3.imread(file_obj, mode='P')  # All frames as gray scale
    flux = np.rot90(np.moveaxis(flux, 0, 2), k=-1, axes=(0, 1))

    meta = {'filename': file_name, '_orig_spatial_wcs': None}
    s3d = Spectrum1D(flux=flux * (u.count / PIX2), meta=standardize_metadata(meta))

    app.add_data(s3d, data_label)
    app.add_data_to_viewer(flux_viewer_reference_name, data_label)


def _get_data_type_by_hdu(hdu):
    # If the data type is some kind of integer, assume it's the mask/dq
    if (hdu.data.dtype in (int, np.uint, np.uint8, np.uint16, np.uint32) or
            any(x in hdu.name.lower() for x in EXT_TYPES['mask'])):
        data_type = 'mask'
    elif ('errtype' in [x.lower() for x in hdu.header.keys()] or
            any(x in hdu.name.lower() for x in EXT_TYPES['uncert'])):
        data_type = 'uncert'
    elif any(x in hdu.name.lower() for x in EXT_TYPES['flux']):
        data_type = 'flux'
    else:
        data_type = ''
    return data_type



import operator
import warnings

import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import NDData
from glue.core.edit_subset_mode import ReplaceMode, OrMode, NewMode
from glue.core.roi import EllipticalROI, CircularROI, CircularAnnulusROI, RectangularROI
from numpy.testing import assert_allclose
from regions import (CircleAnnulusPixelRegion, CirclePixelRegion,
                     CircleSkyRegion, CompoundPixelRegion, CompoundSkyRegion,
                     PixCoord)
from specutils import SpectralRegion

from jdaviz.configs.default.plugins.subset_tools import utils
from jdaviz.core.region_translators import regions2roi


def test_plugin(specviz_helper, spectrum1d):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        specviz_helper.load_data(spectrum1d)
    p = specviz_helper.plugins['Subset Tools']

    # regression test for https://github.com/spacetelescope/jdaviz/issues/1693
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    p.import_region(SpectralRegion(6500 * unit, 7400 * unit))

    p.subset.selected = 'Create New'

    po = specviz_helper.plugins['Plot Options']
    po.layer = 'Subset 1'
    po.line_color = 'green'


def test_subset_definition_with_composite_subset(cubeviz_helper, spectrum1d_cube):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        cubeviz_helper.load_data(spectrum1d_cube)
    cubeviz_helper.app.get_tray_item_from_name('g-subset-tools')


circle_subset_info = {'xc': {'pixel_name': 'X Center (pixels)', 'wcs_name':
                             'RA Center (degrees)', 'initial_value': 5,
                             'final_value': 6},
                      'yc': {'pixel_name': 'Y Center (pixels)', 'wcs_name':
                             'Dec Center (degrees)', 'initial_value': 5,
                             'final_value': 6},
                      'radius': {'pixel_name': 'Radius (pixels)', 'wcs_name':
                                 'Radius (degrees)', 'initial_value': 2,
                                 'final_value': 3}}
elliptical_subset_info = {'xc': {'pixel_name': 'X Center (pixels)', 'wcs_name':
                                 'RA Center (degrees)', 'initial_value': 5.,
                                 'final_value': 6.},
                          'yc': {'pixel_name': 'Y Center (pixels)', 'wcs_name':
                                 'Dec Center (degrees)', 'initial_value': 5.,
                                 'final_value': 6.},
                          'radius_x': {'pixel_name': 'X Radius (pixels)', 'wcs_name':
                                       'RA Radius (degrees)', 'initial_value': 2.,
                                       'final_value': 3.},
                          'radius_y': {'pixel_name': 'Y Radius (pixels)', 'wcs_name':
                                       'Dec Radius (degrees)', 'initial_value': 5.,
                                       'final_value': 6.},
                          'theta': {'pixel_name': 'Angle', 'wcs_name':
                                    'Angle', 'initial_value': 0.,
                                    'final_value': 45.}}
circ_annulus_subset_info = {'xc': {'pixel_name': 'X Center (pixels)', 'wcs_name':
                                   'RA Center (degrees)', 'initial_value': 5.,
                                   'final_value': 6.},
                            'yc': {'pixel_name': 'Y Center (pixels)', 'wcs_name':
                                   'Dec Center (degrees)', 'initial_value': 5.,
                                   'final_value': 6.},
                            'inner_radius': {'pixel_name': 'Inner Radius (pixels)', 'wcs_name':
                                             'Inner Radius (degrees)', 'initial_value': 2.,
                                             'final_value': 3.},
                            'outer_radius': {'pixel_name': 'Outer Radius (pixels)', 'wcs_name':
                                             'Outer Radius (degrees)', 'initial_value': 5.,
                                             'final_value': 6.}}
rectangular_subset_info = {'xmin': {'pixel_name': 'Xmin (pixels)', 'wcs_name':
                                    'RA min (degrees)', 'initial_value': 5.,
                                    'final_value': 6.},
                           'xmax': {'pixel_name': 'Xmax (pixels)', 'wcs_name':
                                    'RA max (degrees)', 'initial_value': 6.,
                                    'final_value': 7.},
                           'ymin': {'pixel_name': 'Ymin (pixels)', 'wcs_name':
                                    'Dec min (degrees)', 'initial_value': 2.,
                                    'final_value': 3.},
                           'ymax': {'pixel_name': 'Ymax (pixels)', 'wcs_name':
                                    'Dec max (degrees)', 'initial_value': 5.,
                                    'final_value': 6.}}


@pytest.mark.parametrize("roi_class, subset_info", [(CircularROI, circle_subset_info),
                                                    (EllipticalROI, elliptical_subset_info),
                                                    (CircularAnnulusROI, circ_annulus_subset_info),
                                                    (RectangularROI, rectangular_subset_info)])
def test_circle_recenter_linking(roi_class, subset_info, imviz_helper, image_2d_wcs):

    arr = np.ones((10, 10))
    ndd = NDData(arr, wcs=image_2d_wcs)
    imviz_helper.load_data(ndd, data_label='dataset1')
    imviz_helper.load_data(ndd, data_label='dataset2')

    # apply subset
    roi_params = {key: subset_info[key]['initial_value'] for key in subset_info}
    imviz_helper.plugins['Subset Tools'].import_region(roi_class(**roi_params))

    # get plugin and check that attribute tracking link type is set properly
    plugin = imviz_helper.plugins['Subset Tools']._obj
    assert not plugin.display_sky_coordinates

    # get initial subset definitions from ROI applied
    subset_defs = plugin.subset_definitions

    # check that the subset definitions, which control what is displayed in the UI, are correct
    for i, attr in enumerate(subset_info):
        assert subset_defs[0][i+1]['name'] == subset_info[attr]['pixel_name']
        assert subset_defs[0][i+1]['value'] == subset_info[attr]['initial_value']

    # get original subset location as a sky region for use later
    original_subs = imviz_helper.app.get_subsets(include_sky_region=True)
    original_sky_region = original_subs['Subset 1'][0]['sky_region']

    # move subset (subset state is what is modified in UI)
    for attr in subset_info:
        plugin._set_value_in_subset_definition(0, subset_info[attr]['pixel_name'],
                                               attr, subset_info[attr]['final_value'])

    # update subset to apply these changes
    plugin.vue_update_subset()
    subset_defs = plugin.subset_definitions

    # and check that it is changed after vue_update_subset runs
    for i, attr in enumerate(subset_info):
        assert subset_defs[0][i+1]['name'] == subset_info[attr]['pixel_name']
        assert subset_defs[0][i+1]['value'] == subset_info[attr]['final_value']

    # get updated subset location as a sky region, we need this later
    updated_sky_region = imviz_helper.app.get_subsets(include_sky_region=True)
    updated_sky_region = updated_sky_region['Subset 1'][0]['sky_region']

    # remove subsets and change link type to wcs
    dc = imviz_helper.app.data_collection
    dc.remove_subset_group(dc.subset_groups[0])
    imviz_helper.link_data(align_by='wcs')
    assert plugin.display_sky_coordinates  # linking change should trigger change to True

    # apply original subset. transform sky coord of original subset to new pixels
    # using wcs of orientation layer (won't be the same as original pixels when pix linked)
    img_wcs = imviz_helper.app.data_collection['Default orientation'].coords
    new_pix_region = original_sky_region.to_pixel(img_wcs)
    new_roi = regions2roi(new_pix_region)
    imviz_helper.plugins['Subset Tools'].import_region(new_roi)

    # get subset definitions again, which should now be in sky coordinates
    subset_defs = plugin.subset_definitions

    # check that the subset definitions, which control what is displayed in the UI,
    # are correct and match original sky region generated when we were first pix linked
    true_values_orig = utils._sky_region_to_subset_def(original_sky_region)

    for i, attr in enumerate(subset_info):
        assert subset_defs[0][i+1]['name'] == subset_info[attr]['wcs_name']
        assert_allclose(subset_defs[0][i+1]['value'], true_values_orig[i]['value'])

    true_values_final = utils._sky_region_to_subset_def(updated_sky_region)

    for i, attr in enumerate(subset_info):
        plugin._set_value_in_subset_definition(0, subset_info[attr]['wcs_name'],
                                               attr, true_values_final[i]['value'])
    # update subset
    plugin.vue_update_subset()

    subset_defs = plugin.subset_definitions
    for i, attr in enumerate(subset_info):
        assert subset_defs[0][i+1]['name'] == subset_info[attr]['wcs_name']
        assert_allclose(subset_defs[0][i+1]['value'], true_values_final[i]['value'])


@pytest.mark.parametrize(
    ('spec_regions', 'mode', 'len_subsets', 'len_subregions'),
    [([SpectralRegion(5.772486091213352 * u.um, 6.052963676101135 * u.um),
       SpectralRegion(6.494371022809778 * u.um, 6.724270682553864 * u.um),
       SpectralRegion(7.004748267441649 * u.um, 7.3404016303483965 * u.um)], 'new', 3, 1),
     ([SpectralRegion(5.772486091213352 * u.um, 6.052963676101135 * u.um),
       SpectralRegion(6.494371022809778 * u.um, 6.724270682553864 * u.um),
       SpectralRegion(7.004748267441649 * u.um, 7.3404016303483965 * u.um)], 'replace', 1, 1),
     ((SpectralRegion(5.772486091213352 * u.um, 6.052963676101135 * u.um) +
       SpectralRegion(6.494371022809778 * u.um, 6.724270682553864 * u.um) +
       SpectralRegion(7.004748267441649 * u.um, 7.3404016303483965 * u.um)), 'or', 1, 3),
     ((SpectralRegion(5.772486091213352 * u.um, 6.052963676101135 * u.um) +
       SpectralRegion(5.8 * u.um, 5.9 * u.um) +
       SpectralRegion(6.494371022809778 * u.um, 6.724270682553864 * u.um) +
       SpectralRegion(7 * u.um, 7.2 * u.um)), ['new', 'andnot', 'or', 'or'], 1, 4),
     (SpectralRegion(5.8 * u.um, 5.9 * u.um), None, 1, 1)
     ]
)
def test_import_spectral_region(cubeviz_helper, spectrum1d_cube, spec_regions, mode, len_subsets,
                                len_subregions):
    cubeviz_helper.load_data(spectrum1d_cube)
    plg = cubeviz_helper.plugins['Subset Tools']
    plg.import_region(spec_regions, combination_mode=mode)
    subsets = cubeviz_helper.app.get_subsets()
    assert len(subsets) == len_subsets
    assert len(subsets['Subset 1']) == len_subregions
    assert cubeviz_helper.app.session.edit_subset_mode.mode == ReplaceMode


def test_import_spectral_regions_file(cubeviz_helper, spectrum1d_cube, tmp_path):
    cubeviz_helper.load_data(spectrum1d_cube)
    plg = cubeviz_helper.plugins['Subset Tools']
    s = SpectralRegion(5*u.um, 6*u.um)
    local_path = str(tmp_path / 'spectral_region.ecsv')
    s.write(local_path)
    plg.import_region(local_path)
    subsets = cubeviz_helper.app.get_subsets()
    assert len(subsets) == 1

    plg.combination_mode = 'or'
    plg.import_region(SpectralRegion(7 * u.um, 8 * u.um))

    subsets = cubeviz_helper.app.get_subsets()
    assert len(subsets['Subset 1']) == 2

    subset2 = (SpectralRegion(5.772486091213352 * u.um, 6.052963676101135 * u.um) +
               SpectralRegion(5.8 * u.um, 5.9 * u.um))
    plg.import_region(subset2, combination_mode=['new', 'andnot'])

    assert cubeviz_helper.app.session.edit_subset_mode.mode == OrMode

    with pytest.raises(ValueError, match='\'test\' not one of'):
        plg.combination_mode = 'test'


def test_get_regions(cubeviz_helper, spectrum1d_cube, imviz_helper):

    """Test Subset Tools.get regions."""

    cubeviz_helper.load_data(spectrum1d_cube)
    plg = cubeviz_helper.plugins['Subset Tools']

    # load one spectral region, which will become 'Subset 1'
    plg.import_region(SpectralRegion(1 * u.um, 2 * u.um))

    # load one spatial region, which will become 'Subset 2'
    spatial_reg = CirclePixelRegion(center=PixCoord(x=2, y=2), radius=2)
    plg.import_region(spatial_reg, combination_mode='new')

    # call get_regions, which by default for cubeviz will return both
    # spatial and spectral regions
    all_regions = plg.get_regions()
    assert len(all_regions) == 2

    # make sure filtering by subset label works
    only_s1 = plg.get_regions(list_of_subset_labels=['Subset 1'])
    assert len(only_s1) == 1
    assert only_s1['Subset 1']

    # now specify region type and check output
    spatial_regions = plg.get_regions(region_type='spatial')
    assert len(spatial_regions) == 1
    assert spatial_regions['Subset 2']
    spectral_regions = plg.get_regions(region_type='spectral')
    assert len(spectral_regions) == 1
    assert spectral_regions['Subset 1']

    # now test a composite spatial subset, make sure it is retrieved
    # as a CompoundRegion (since operator is &)
    sr1 = CirclePixelRegion(center=PixCoord(x=2.5, y=2.5), radius=2)
    sr2 = CirclePixelRegion(center=PixCoord(x=2.5, y=3), radius=2)
    plg.import_region(sr1, combination_mode='new')
    plg.import_region(sr2, combination_mode='and')
    spatial_regions = plg.get_regions(region_type='spatial')
    assert spatial_regions['Subset 3']

    # test errors
    with pytest.raises(ValueError, match='No spectral subests in imviz.'):
        imviz_helper.plugins['Subset Tools'].get_regions('spectral')
    with pytest.raises(ValueError, match="`region_type` must be 'spectral', 'spatial', or None for any."):  # noqa E501
        plg.get_regions(region_type='fail')


def test_get_regions_composite(imviz_helper):
    """
    Test the behavior of retrieving composite subsets as regions
    using subset_tools.get_regions
    """
    a = np.ones((200, 200))
    imviz_helper.load_data(a, data_label="test")
    plg = imviz_helper.plugins['Subset Tools']

    # apply two (not concentric) circular subsets
    plg.import_region(CirclePixelRegion(center=PixCoord(x=96.0, y=96.0),
                                        radius=45.0), combination_mode='new')
    plg.import_region(CirclePixelRegion(center=PixCoord(x=100.0, y=100.0),
                                        radius=25.0), combination_mode='new')

    # create a composite subset created from these circular subsets
    # that are applied with remove/and not to subtract the smaller from the larger
    imviz_helper.app.session.edit_subset_mode._mode = NewMode
    subset_groups = imviz_helper.app.data_collection.subset_groups
    new_subset = subset_groups[0].subset_state & ~subset_groups[1].subset_state
    imviz_helper.default_viewer._obj.apply_subset_state(new_subset)

    # call get_regions and make sure Subset 3, the composite subset, is retrieved,
    # but as a list of tuples of  individual subsets and their combination modes,
    # since they were created with &~ and do not form an exact circular annulus
    regions = plg.get_regions()
    assert sorted(regions) == ["Subset 1", "Subset 2", "Subset 3"]
    assert len(regions['Subset 3']) == 2
    assert isinstance(regions['Subset 3'][0][0], CirclePixelRegion)
    assert regions['Subset 3'][0][1] == ''  # first operator should be ''
    assert isinstance(regions['Subset 3'][1][0], CirclePixelRegion)
    assert regions['Subset 3'][1][1] == 'AndNotState'

    # make sure the same regions are returned by app.get_subsets
    get_subsets = imviz_helper.app.get_subsets()
    assert sorted(get_subsets) == ["Subset 1", "Subset 2", "Subset 3"]

    # Now, create two concentric circular subsets and combine them to form a circular annulus.
    plg.import_region(CirclePixelRegion(center=PixCoord(x=95.0, y=95.0),
                                        radius=45.0), combination_mode='new')
    plg.import_region(CirclePixelRegion(center=PixCoord(x=95.0, y=95.0),
                                        radius=25.0), combination_mode='new')
    imviz_helper.app.session.edit_subset_mode._mode = NewMode
    subset_groups = imviz_helper.app.data_collection.subset_groups
    new_subset = subset_groups[3].subset_state & ~subset_groups[4].subset_state
    imviz_helper.default_viewer._obj.apply_subset_state(new_subset)

    # now, when get_regions is called, the combined subset should be represented as a
    # CircleAnnulusPixelRegion rather than two circlular subsets
    regions = plg.get_regions()
    assert isinstance(regions['Subset 6'], CircleAnnulusPixelRegion)

    # If and/or/xor are used to create composite subsets, they can be
    # represented as CompoundRegions.
    for sub, op in [('7', operator.and_), ('8', operator.or_), ('9', operator.xor)]:
        imviz_helper.app.session.edit_subset_mode._mode = NewMode
        new_subset = op(subset_groups[3].subset_state, subset_groups[4].subset_state)
        imviz_helper.default_viewer._obj.apply_subset_state(new_subset)
        regions = plg.get_regions()
        assert isinstance(regions[f'Subset {sub}'], CompoundPixelRegion)


def test_get_regions_composite_wcs_linked(imviz_helper, image_2d_wcs):
    data = NDData(np.ones((128, 128)) * u.nJy, wcs=image_2d_wcs)
    imviz_helper.load_data(data)

    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    st = imviz_helper.plugins['Subset Tools']

    sr1 = CircleSkyRegion(center=SkyCoord(ra=337.5058778*u.deg,
                          dec=-20.808486*u.deg), radius=0.008*u.deg)
    sr2 = CircleSkyRegion(center=SkyCoord(ra=337.51*u.deg, dec=-20.81*u.deg),
                          radius=0.007*u.deg)
    st.import_region(sr1, combination_mode='new')
    st.import_region(sr2, combination_mode='and')

    # composite subset should be a sky region, and combined to a compound region
    regs = st.get_regions()

    cr = regs['Subset 1']
    assert isinstance(cr, CompoundSkyRegion)
    assert_allclose(cr.region1.center.ra.deg, sr1.center.ra.deg)
    assert_allclose(cr.region2.center.ra.deg, sr2.center.ra.deg)
    assert cr.operator == operator.and_


@pytest.mark.skip(reason="Unskip after JDAT-5186.")
def test_get_composite_sky_region_remove(imviz_helper, image_2d_wcs):
    """
    Test to ensure bug fixed by JDAT-5186 is fixed, where get_subsets
    for compoisute subset applied with 'remove' when WCS linked was not
    correctly retrieving the second subset.
    """
    data = NDData(np.ones((128, 128)) * u.nJy, wcs=image_2d_wcs)
    imviz_helper.load_data(data)

    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    st = imviz_helper.plugins['Subset Tools']

    sr1 = CircleSkyRegion(center=SkyCoord(ra=337.5058778*u.deg,
                          dec=-20.808486*u.deg), radius=0.008*u.deg)
    sr2 = CircleSkyRegion(center=SkyCoord(ra=337.51*u.deg, dec=-20.81*u.deg),
                          radius=0.007*u.deg)
    st.import_region(sr1, combination_mode='new')
    st.import_region(sr2, combination_mode='andnot')

    # app.get_subsets
    ss = imviz_helper.app.get_subsets(include_sky_region=True)

    # composite region should be returned as 2 sky regions
    assert isinstance(ss['Subset 1'][0]['sky_region'], CircleSkyRegion)
    assert isinstance(ss['Subset 1'][1]['sky_region'], CircleSkyRegion)

    # now make sure Subset Tools get_regions agrees and both sky regions
    # are returned for Subset 1
    regs = st.get_regions()
    assert isinstance(regs['Subset 1'][0], CircleSkyRegion)
    assert isinstance(regs['Subset 1'][1], CircleSkyRegion)


def test_check_valid_subset_label(imviz_helper):

    # imviz instance with some data
    data = NDData(np.ones((50, 50)) * u.nJy)
    imviz_helper.load_data(data)

    st = imviz_helper.plugins["Subset Tools"]

    # apply three subsets, with their default names of `Subset 1`, `Subset 2`, and `Subset 3`
    st.import_region(CircularROI(20, 20, 10))
    st.subset = "Create New"
    st.import_region(CircularROI(25, 25, 10))
    st.subset = "Create New"
    st.import_region(CircularROI(30, 30, 10))

    # we should not be able to rename or add a subset named 'Subset 1'.
    # Make sure this warns and returns accordingly.
    with pytest.raises(ValueError, match="Cannot rename subset to name of an existing subset"):
        st.rename_selected("Subset 1")

    with pytest.raises(ValueError, match="The pattern 'Subset N' is reserved"):
        st.rename_selected("Subset 5")


def test_rename_subset(cubeviz_helper, spectrum1d_cube):

    cubeviz_helper.load_data(spectrum1d_cube)
    plg = cubeviz_helper.plugins['Subset Tools']

    spatial_reg = CirclePixelRegion(center=PixCoord(x=2, y=2), radius=2)
    plg.import_region(spatial_reg, combination_mode='new')
    spatial_reg = CirclePixelRegion(center=PixCoord(x=4, y=4), radius=1)
    plg.import_region(spatial_reg, combination_mode='new')

    plg.rename_subset("Subset 1", "Test Rename")

    assert plg.subset.choices == ['Create New', 'Test Rename', 'Subset 2']
    assert cubeviz_helper.app.data_collection[-1].label == "Spectrum (Test Rename, sum)"

    plg.rename_selected("Second Test")
    assert plg.subset.choices == ['Create New', 'Test Rename', 'Second Test']

    with pytest.raises(ValueError, match="No subset named BadLabel to rename"):
        plg.rename_subset("BadLabel", "Failure")


def test_update_subset(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube)
    plg = cubeviz_helper.plugins['Subset Tools']

    spatial_reg = CirclePixelRegion(center=PixCoord(x=2, y=2), radius=2)
    plg.import_region(spatial_reg, combination_mode='new')
    spatial_reg = CirclePixelRegion(center=PixCoord(x=4, y=4), radius=1)
    plg.import_region(spatial_reg, combination_mode='and')

    subset_def = plg.update_subset('Subset 1')
    print(subset_def)
    assert isinstance(subset_def, dict)
    assert len(subset_def) == 2

    with pytest.raises(ValueError, match='subset has more than one subregion'):
        plg.update_subset('Subset 1', xc=1)

    with pytest.raises(ValueError, match='not an attribute of the specified subset/subregion.'):
        plg.update_subset('Subset 1', subregion=0, notanattribute=1)

    plg.update_subset('Subset 1', subregion=0, xc=3, yc=1, radius=1)

    # Check xc
    assert plg._obj.subset_definitions[0][1]['value'] == 3
    # Check yc
    assert plg._obj.subset_definitions[0][2]['value'] == 1
    # Check radius
    assert plg._obj.subset_definitions[0][3]['value'] == 1



from glue_jupyter.bqplot.profile import BqplotProfileView

from jdaviz.core.tools import SinglePixelRegion
from jdaviz.core.marks import PluginLine

__all__ = ['ProfileFromCube']


class ProfileFromCube(SinglePixelRegion):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._profile_viewer = None
        self._previous_bounds = None
        self._mark = None
        self._data = None

    def _reset_profile_viewer_bounds(self):
        self._profile_viewer.set_limits(
            y_min=self._previous_bounds[2], y_max=self._previous_bounds[3])

    def activate(self):
        self.viewer.add_event_callback(self.on_mouse_move, events=['mousemove', 'mouseleave'])
        if self._profile_viewer is None:
            # Get first profile viewer
            for _, viewer in self.viewer.jdaviz_helper.app._viewer_store.items():
                if isinstance(viewer, BqplotProfileView):
                    self._profile_viewer = viewer
                    break
        if self._mark is None:
            self._mark = PluginLine(self._profile_viewer, visible=False)
            self._profile_viewer.figure.marks = self._profile_viewer.figure.marks + [self._mark, ]
        # Store these so we can revert to previous user-set zoom after preview view
        self._previous_bounds = self._profile_viewer.get_limits()
        super().activate()

    def deactivate(self):
        self.viewer.remove_event_callback(self.on_mouse_move)
        self._reset_profile_viewer_bounds()
        super().deactivate()

    def on_mouse_move(self, data):
        raise NotImplementedError("must be implemented by subclasses")



from traitlets import List

from jdaviz.core.events import NewViewerMessage
from jdaviz.core.registries import tool_registry, viewer_registry
from jdaviz.core.template_mixin import TemplateMixin

__all__ = ['ViewerCreator']


@tool_registry('g-viewer-creator')
class ViewerCreator(TemplateMixin):
    template_file = __file__, "viewer_creator.vue"
    viewer_types = List([]).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load in the references to the viewer registry. Because traitlets
        #  can't serialize the actual viewer class reference, create a list of
        #  dicts containing just the viewer name and label.
        self.viewer_types = [{'name': k, 'label': v['label']}
                             for k, v in viewer_registry.members.items()]

    def vue_create_viewer(self, name):
        viewer_cls = viewer_registry.members[name]['cls']

        # selected = self.components.get('g-data-tree').selected

        # for idx in selected:
        #     data = validate_data_argument(self.data_collection,
        #                                   self.data_collection[idx])

        new_viewer_message = NewViewerMessage(
            viewer_cls, data=None, sender=self)

        self.hub.broadcast(new_viewer_message)



import numpy as np

from astropy import units as u
from bqplot import LinearScale
from bqplot.marks import Lines, Label, Scatter
from glue.core import HubListener
from specutils import Spectrum1D

from jdaviz.core.events import GlobalDisplayUnitChanged
from jdaviz.core.events import (SliceToolStateMessage, LineIdentifyMessage,
                                SpectralMarksChangedMessage,
                                RedshiftMessage)
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               check_if_unit_is_per_solid_angle,
                                               flux_conversion_general)


__all__ = ['OffscreenLinesMarks', 'BaseSpectrumVerticalLine', 'SpectralLine',
           'SliceIndicatorMarks', 'ShadowMixin', 'ShadowLine', 'ShadowLabelFixedY',
           'PluginMark', 'LinesAutoUnit', 'PluginLine', 'PluginScatter',
           'LineAnalysisContinuum', 'LineAnalysisContinuumCenter',
           'LineAnalysisContinuumLeft', 'LineAnalysisContinuumRight',
           'LineUncertainties', 'ScatterMask', 'SelectedSpaxel', 'MarkersMark',
           'CatalogMark', 'FootprintOverlay', 'ApertureMark']

accent_color = "#c75d2c"


class OffscreenLinesMarks(HubListener):
    def __init__(self, viewer):
        self.viewer = viewer

        viewer.state.add_callback("x_min", lambda x_min: self._update_counts())
        viewer.state.add_callback("x_max", lambda x_max: self._update_counts())

        viewer.session.hub.subscribe(self, RedshiftMessage,
                                     handler=self._update_counts)
        viewer.session.hub.subscribe(self, SpectralMarksChangedMessage,
                                     handler=self._update_counts)

        self.left = Label(text=[''], x=[0.02], y=[0.8],
                          scales={'x': LinearScale(min=0, max=1),
                                  'y': LinearScale(min=0, max=1)},
                          colors=['gray'], default_size=12,
                          align='start')
        self.right = Label(text=[''], x=[0.98], y=[0.8],
                           scales={'x': LinearScale(min=0, max=1),
                                   'y': LinearScale(min=0, max=1)},
                           colors=['gray'], default_size=12,
                           align='end')

        self._update_counts()

    @property
    def marks(self):
        return [self.left, self.right]

    def _update_counts(self, *args):
        oob_left, oob_right = 0, 0
        for m in self.viewer.figure.marks:
            if isinstance(m, SpectralLine):
                if m.x[0] < self.viewer.state.x_min:
                    oob_left += 1
                elif m.x[0] > self.viewer.state.x_max:
                    oob_right += 1
        self.left.text = [f'\u25c0 {oob_left}' if oob_left > 0 else '']
        self.right.text = [f'{oob_right} \u25b6' if oob_right > 0 else '']


class PluginMark:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.xunit = None
        self.yunit = None
        # whether to update existing marks when global display units are changed
        self.auto_update_units = True
        self.hub.subscribe(self, GlobalDisplayUnitChanged,
                           handler=self._on_global_display_unit_changed)

        if self.xunit is None:
            self.set_x_unit()
        if self.yunit is None:
            self.set_y_unit()

    @property
    def hub(self):
        return self.viewer.hub

    def update_xy(self, x, y):
        # If x and y are not in the previous units, they should be provided as quantities
        if hasattr(x, 'value'):
            xunit = x.unit
            x = x.value
        else:
            xunit = None
        self.x = np.asarray(x)
        if xunit is not None:
            self.xunit = u.Unit(xunit)

        if hasattr(y, 'value'):
            yunit = y.unit
            y = y.value
        else:
            yunit = None
        self.y = np.asarray(y)
        if yunit is not None:
            self.yunit = u.Unit(yunit)

    def append_xy(self, x, y):
        self.x = np.append(self.x, x)
        self.y = np.append(self.y, y)

    def set_x_unit(self, unit=None):
        if unit is None:
            if not hasattr(self.viewer.state, 'x_display_unit'):
                return
            unit = self.viewer.state.x_display_unit
        unit = u.Unit(unit)

        if self.xunit is not None and not np.all([s == 0 for s in self.x.shape]):
            x = (self.x * self.xunit).to_value(unit, u.spectral())
            self.xunit = unit
            self.x = x
        self.xunit = unit

    def set_y_unit(self, unit=None):
        if unit is None:
            if not hasattr(self.viewer.state, 'y_display_unit'):
                return
            unit = self.viewer.state.y_display_unit
        unit = u.Unit(unit)

        # spectrum y-values in viewer have already been converted, don't convert again
        # if a spectral_y_type is changed, just update the unit
        if self.yunit is not None and check_if_unit_is_per_solid_angle(self.yunit) != check_if_unit_is_per_solid_angle(unit):  # noqa
            self.yunit = unit
            return

        if self.yunit is not None and not np.all([s == 0 for s in self.y.shape]):  # noqa

            if self.viewer.default_class is Spectrum1D:
                if self.xunit is None:
                    return
                spec = self.viewer.state.reference_data.get_object(cls=Spectrum1D)

                pixar_sr = spec.meta.get('PIXAR_SR', 1)
                cube_wave = self.x * self.xunit
                equivs = all_flux_unit_conversion_equivs(pixar_sr, cube_wave)
                y = flux_conversion_general(self.y, self.yunit, unit,
                                            equivs, with_unit=False)
            self.y = y

        self.yunit = unit

    def _on_global_display_unit_changed(self, msg):
        if not self.auto_update_units:
            return
        if self.viewer.__class__.__name__ in ['SpecvizProfileView',
                                              'CubevizProfileView',
                                              'MosvizProfileView',
                                              'MosvizProfile2DView']:
            axis_map = {'spectral': 'x', 'spectral_y': 'y'}
        else:
            return
        axis = axis_map.get(msg.axis, None)
        if axis is not None:
            scale = self.scales.get(axis, None)
            # if PluginMark mark is LinearScale(0, 1), prevent it from entering unit conversion
            # machinery so it maintains it's position in viewer.
            if isinstance(scale, LinearScale) and (scale.min, scale.max) == (0, 1):
                return

            getattr(self, f'set_{axis}_unit')(msg.unit)

    def clear(self):
        self.update_xy([], [])


class BaseSpectrumVerticalLine(Lines, PluginMark, HubListener):
    def __init__(self, viewer, x, **kwargs):
        self.viewer = viewer

        # the location of the marker will need to update automatically if the
        # underlying data changes (through a unit conversion, for example)
        if hasattr(viewer.state, 'reference_data'):
            viewer.state.add_callback("reference_data",
                                      self._update_reference_data)

        scales = viewer.scales

        # Lines.__init__ will set self.x
        super().__init__(x=[x, x], y=[0, 1],
                         scales={'x': scales['x'], 'y': LinearScale(min=0, max=1)},
                         **kwargs)

    def _update_reference_data(self, reference_data):
        # don't update x units before initialization or in rampviz
        if reference_data is None or self.viewer.jdaviz_app.config == 'rampviz':
            return

        self._update_unit(reference_data.get_object(cls=Spectrum1D).spectral_axis.unit)

    def _update_unit(self, new_unit):
        # the x-units may have changed.  We want to convert the internal self.x
        # from self.xunit to the new units (x_all.unit)
        if self.xunit is None:
            self.xunit = new_unit
            return
        if new_unit == self.xunit:
            return
        old_quant = self.x[0]*self.xunit
        x = old_quant.to_value(new_unit, equivalencies=u.spectral())
        self.x = [x, x]
        self.xunit = new_unit


class SpectralLine(BaseSpectrumVerticalLine):
    """
    Subclass on bqplot Lines, mostly so that we can erase spectral lines
    by eliminating any SpectralLines objects from a figures marks list. Also
    lets us do wavelength redshifting here on mark creation.
    """
    def __init__(self, viewer, rest_value, redshift=0, name=None, **kwargs):
        self._rest_value = rest_value
        self._identify = False
        self.name = name

        # table_index is same as name_rest elsewhere
        self.table_index = kwargs.pop("table_index", None)

        # setting redshift will set self.x and enable the obs_value property,
        # but to do that we need x_unit set first (would normally be assigned
        # in the super init)
        self.xunit = u.Unit(viewer.state.x_display_unit)
        self.redshift = redshift

        viewer.session.hub.subscribe(self, LineIdentifyMessage,
                                     handler=self._process_identify_change)

        super().__init__(viewer=viewer, x=self.obs_value, stroke_width=1,
                         fill='none', close_path=False, **kwargs)

    @property
    def name_rest(self):
        return self.table_index

    @property
    def rest_value(self):
        return self._rest_value

    @property
    def obs_value(self):
        return self.x[0]

    def set_x_unit(self, unit=None):
        prev_unit = self.xunit
        super().set_x_unit(unit=unit)
        self._rest_value = (self._rest_value * prev_unit).to_value(unit, u.spectral())

    @property
    def redshift(self):
        return self._redshift

    @redshift.setter
    def redshift(self, redshift):
        self._redshift = redshift
        if str(self.xunit.physical_type) == 'length':
            obs_value = self._rest_value*(1+redshift)
        elif str(self.xunit.physical_type) == 'frequency':
            obs_value = self._rest_value/(1+redshift)
        else:
            # catch all for anything else (wavenumber, energy, etc)
            rest_angstrom = (self._rest_value*self.xunit).to_value(u.Angstrom,
                                                                   equivalencies=u.spectral())
            obs_angstrom = rest_angstrom*(1+redshift)
            obs_value = (obs_angstrom*u.Angstrom).to_value(self.xunit,
                                                           equivalencies=u.spectral())
        self.x = [obs_value, obs_value]

    @property
    def identify(self):
        return self._identify

    @identify.setter
    def identify(self, identify):
        if not isinstance(identify, bool):  # pragma: no cover
            raise TypeError("identify must be of type bool")

        self._identify = identify
        self.stroke_width = 3 if identify else 1

    def _process_identify_change(self, msg):
        self.identify = msg.name_rest == self.table_index

    def _update_unit(self, new_unit):
        if self.xunit is None:
            self.xunit = new_unit
            return
        if new_unit == self.xunit:
            return

        old_quant = self._rest_value*self.xunit
        self._rest_value = old_quant.to_value(new_unit, equivalencies=u.spectral())
        # re-compute self.x from current redshift (instead of converting that as well)
        self.redshift = self._redshift
        self.xunit = new_unit


class SliceIndicatorMarks(BaseSpectrumVerticalLine, HubListener):
    """Subclass on bqplot Lines to handle slice/wavelength indicator.
    """
    def __init__(self, viewer, value=0, **kwargs):
        self._viewer = viewer
        self._value = None
        self._oob = False  # out-of-bounds, either False, 'left', or 'right'
        self._active = False
        # TODO: new viewers need to respect plugin settings
        self._show_if_inactive = True
        self._show_value = True

        viewer.state.add_callback("x_min", lambda x_min: self._value_handle_oob(update_label=True))
        viewer.state.add_callback("x_max", lambda x_max: self._value_handle_oob(update_label=True))
        viewer.session.hub.subscribe(self, SliceToolStateMessage,
                                     handler=self._on_change_state)

        super().__init__(viewer=viewer,
                         x=[value, value],
                         stroke_width=2,
                         marker='diamond',
                         fill='none', close_path=False,
                         labels=['slice'], labels_visibility='none', **kwargs)

        self.value = value

        # instead of using the Lines label which is limited, we'll use a Label object which
        # will follow the x-coordinate of the slice indicator line, with a fixed y-value
        # (in axes-units) and will flip its alignment depending on whether the line is on the
        # left or right side of the axes.
        self.label = ShadowLabelFixedY(viewer, self, shadow_traits=[], default_size=12, y=0.95)

        # default to the initial state of the tool since we can't control if this will
        # happen before or after the initialization of the tool
        tool_active = self.viewer.toolbar.active_tool_id == 'jdaviz:selectslice'
        self._on_change_state({'active': tool_active})

    @property
    def marks(self):
        return [self, self.label]

    def _on_global_display_unit_changed(self, msg):
        # Updating the value is handled by the plugin itself, need to update unit string.
        if msg.axis in ["spectral", "x"]:
            self.xunit = msg.unit
        self._update_label()

    def _value_handle_oob(self, x=None, update_label=False):
        if x is None:
            x = self.value
        else:
            self._value = x
        x_min, x_max = self._viewer.state.x_min, self._viewer.state.x_max
        if x_min is None or x_max is None:
            self.x = [x, x]
            return
        x_range = x_max - x_min
        padding_fig = 0.01
        padding = padding_fig * x_range
        x_min += padding
        x_max -= padding
        # ensure y-scale has been set (we'll only be overriding x, but scatter viewers complain
        # if y-scale is not set)
        self.scales.setdefault('y', LinearScale(min=0, max=1))
        if x < x_min:
            self.x = [padding_fig, padding_fig]
            self.scales = {**self.scales, 'x': LinearScale(min=0, max=1)}
            self.line_style = 'dashed'
            self._oob = 'left'
        elif x > x_max:
            self.x = [1-padding_fig, 1-padding_fig]
            self.scales = {**self.scales, 'x': LinearScale(min=0, max=1)}
            self.line_style = 'dashed'
            self._oob = 'right'
        else:
            self.x = [x, x]
            self.scales = {**self.scales, 'x': self._viewer.scales['x']}
            self.line_style = 'solid'
            self._oob = False
        if update_label:
            self._update_label()

    def _update_colors_opacities(self):
        # orange (accent) if active, import button blue otherwise (see css in main_styles.vue)
        if not self._show_if_inactive and not self._active:
            self.label.visible = False
            self.visible = False
            return

        self.visible = True
        self.label.visible = self._show_value
        self.colors = ["#c75109" if self._active else "#007BA1"]
        self.opacities = [1.0 if self._active else 0.9]

    def _on_change_state(self, msg={}):
        if isinstance(msg, dict):
            changes = msg
        else:
            if msg.viewer is not None and msg.viewer != self.viewer:
                return
            changes = msg.change

        for k, v in changes.items():
            if k == 'active':
                self._active = v
            elif k == 'show_indicator':
                self._show_if_inactive = v
            elif k == 'show_value':
                self._show_value = v

        self._update_colors_opacities()

    def _update_label(self):
        def _formatted_value(value):
            power = abs(np.log10(value))
            if power >= 3:
                # use scientific notation
                return f'{value:0.4e}'
            else:
                return f'{value:0.4f}'

        valuestr = _formatted_value(self.value)
        xunit = str(self.xunit) if self.xunit is not None else ''
        # U+00A0 is a blank space, U+25C0 a left arrow triangle, and U+25B6 a right arrow triangle
        if self._oob == 'left':
            self.labels = [f'\u00A0 \u25c0 {valuestr} {xunit} \u00A0']  # noqa
        elif self._oob == 'right':
            self.labels = [f'{valuestr} {xunit} \u25b6 \u00A0']
        else:
            self.labels = [f'\u00A0 {valuestr} {xunit} \u00A0']

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value_handle_oob(value, update_label=True)


class ShadowMixin:
    """Mixin class to propagate traits from one mark object to another.
    Anything in ``sync_traits`` will be mirrored directly from
    ``shadowing`` to the shadowed object.

    Can manually override ``_on_shadowing_changed`` for more advanced logic cases.
    """
    def _get_id(self, mark):
        return getattr(mark, '_model_id', None)

    def _setup_shadowing(self, shadowing, sync_traits=[], other_traits=[]):
        """
        sync_traits: traits to set now, and mirror any changes to shadowing in the future
        other_trait: traits to set now, but not mirror in the future
        """
        if not hasattr(self, '_shadowing'):
            self._shadowing = {}
            self._sync_traits = {}
        shadowing_id = self._get_id(shadowing)
        self._shadowing[shadowing_id] = shadowing
        self._sync_traits[shadowing_id] = sync_traits

        # sync initial values
        for attr in sync_traits + other_traits:
            self._on_shadowing_changed({'name': attr,
                                        'new': getattr(shadowing, attr),
                                        'owner': shadowing})

        # subscribe to future changes
        shadowing.observe(self._on_shadowing_changed)

    def _on_shadowing_changed(self, change):
        if change['name'] in self._sync_traits.get(self._get_id(change.get('owner')), []):
            setattr(self, change['name'], change['new'])
        return


class ShadowLine(Lines, HubListener, ShadowMixin):
    """Create a white shadow line around another line
    to help make it standout on top of other lines.
    """
    def __init__(self, shadowing, shadow_width=1, **kwargs):
        self._shadow_width = shadow_width
        super().__init__(scales=shadowing.scales,
                         stroke_width=shadowing.stroke_width+shadow_width if shadowing.stroke_width else 0, # noqa
                         marker_size=shadowing.marker_size+shadow_width if shadowing.marker_size else 0, # noqa
                         colors=[kwargs.pop('color', 'white')],
                         **kwargs)

        self._setup_shadowing(shadowing,
                              ['scales', 'x', 'y', 'visible', 'line_style', 'marker'],
                              ['stroke_width', 'marker_size'])


class ShadowLabelFixedY(Label, ShadowMixin):
    """Label whose position shadows that of a parent ``shadowing``
    line and will flip alignment based on whether it is left or
    right of the center of the viewer.
    """
    def __init__(self, viewer, shadowing, shadow_traits=['visible'],
                 y=0.95, point_index=0, **kwargs):
        super().__init__(**kwargs)
        self._viewer = viewer
        self.y = [y]
        self.scales['y'] = LinearScale(min=0, max=1)
        self._point_index = point_index
        self._setup_shadowing(shadowing,
                              shadow_traits,
                              ['x', 'scales', 'labels', 'colors'])

        viewer.state.add_callback("x_min", lambda x_min: self._update_align())
        viewer.state.add_callback("x_max", lambda x_max: self._update_align())

    def _force_redraw(self):
        # TODO: bug in bqplot that change in align/colors traitlet doesn't update immediately,
        # we'll get around it in the meantime by just forcing the Label to see a change to the
        # text traitlet
        text = self.text
        self.text = ['']
        self.text = text

    def _update_align(self):
        if not isinstance(self.scales.get('x'), LinearScale):
            return
        # determine alignment automatically
        if self.scales['x'].min == 0 and self.scales['x'].max == 1:
            # then we're in axes units, so just check position compared to 0.5
            is_to_right = self.x[0] > 0.5
        else:
            # then we're in data units, so check position compared to the median of the axes limits
            is_to_right = self.x[0] > (self._viewer.state.x_min + self._viewer.state.x_max) / 2.

        if is_to_right and self.align != 'end':
            self.align = 'end'
            # force redraw by re-updating label
            self._force_redraw()
        if not is_to_right and self.align != 'start':
            self.align = 'start'
            # force redraw by re-updating label
            self._force_redraw()

    def _on_shadowing_changed(self, change):
        super()._on_shadowing_changed(change)
        if change['name'] == 'labels':
            self.text = [change['new'][self._point_index]]
        elif change['name'] in ('x', 'colors'):
            setattr(self, change['name'], [change['new'][self._point_index]])
            if change['name'] == 'colors':
                # bqplot bug that won't notice change to colors, manually force re-draw
                self._force_redraw()
        elif change['name'] == 'scales':
            self.scales = {**self.scales, 'x': change['new']['x']}

        if change['name'] in ('x', 'scales'):
            # then the position of the label on the plot has changed, so re-determine whether
            # it should be aligned to the left or right
            self._update_align()


class LinesAutoUnit(PluginMark, Lines, HubListener):
    def __init__(self, viewer, *args, **kwargs):
        self.viewer = viewer
        super().__init__(*args, **kwargs)


class PluginLine(Lines, PluginMark, HubListener):
    def __init__(self, viewer, x=[], y=[], **kwargs):
        self.viewer = viewer
        # color is same blue as import button
        kwargs.setdefault('colors', [accent_color])
        self.label = kwargs.get('label')
        super().__init__(x=x, y=y, scales=kwargs.pop('scales', viewer.scales), **kwargs)


class PluginScatter(Scatter, PluginMark, HubListener):
    def __init__(self, viewer, x=[], y=[], **kwargs):
        self.viewer = viewer
        # default color is same blue as import button
        kwargs.setdefault('colors', [accent_color])
        super().__init__(x=x, y=y, scales=kwargs.pop('scales', viewer.scales), **kwargs)


class LineAnalysisContinuum(PluginLine):
    def __init__(self, *args, **kwargs):
        # units do not need to be updated because the plugin itself reruns
        # the computation and automatically changes the arrays themselves
        self.auto_update_units = kwargs.pop('auto_update_units', False)
        super().__init__(*args, **kwargs)


class LineAnalysisContinuumCenter(LineAnalysisContinuum):
    def __init__(self, viewer, x=[], y=[], **kwargs):
        super().__init__(viewer, x, y, **kwargs)
        self.stroke_width = 1


class LineAnalysisContinuumLeft(LineAnalysisContinuum):
    def __init__(self, viewer, x=[], y=[], **kwargs):
        super().__init__(viewer, x, y, **kwargs)
        self.stroke_width = 5


class LineAnalysisContinuumRight(LineAnalysisContinuumLeft):
    pass


class LineUncertainties(LinesAutoUnit):
    def __init__(self, viewer, *args, **kwargs):
        super().__init__(viewer, *args, **kwargs)


class ScatterMask(Scatter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SelectedSpaxel(Lines):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class MarkersMark(PluginScatter):
    def __init__(self, viewer, **kwargs):
        kwargs.setdefault('marker', 'circle')
        super().__init__(viewer, **kwargs)


class CatalogMark(PluginScatter):
    def __init__(self, viewer, **kwargs):
        kwargs.setdefault('marker', 'circle')
        super().__init__(viewer, **kwargs)


class FootprintOverlay(PluginLine):
    def __init__(self, viewer, overlay, **kwargs):
        self._overlay = overlay
        kwargs.setdefault('stroke_width', 2)
        kwargs.setdefault('close_path', True)
        kwargs.setdefault('opacities', [0.8])
        kwargs.setdefault('fill', 'inside')
        kwargs.setdefault('fill_opacities', [0.2])
        super().__init__(viewer, **kwargs)

    @property
    def overlay(self):
        return self._overlay


class ApertureMark(PluginLine):
    def __init__(self, viewer, id, **kwargs):
        self._id = id
        super().__init__(viewer, **kwargs)


class HistogramMark(Lines):
    def __init__(self, min_max_value, scales, **kwargs):
        # Vertical line in LinearScale
        y = [0, 1]
        colors = [accent_color]
        line_style = "solid"
        super().__init__(x=min_max_value, y=y, scales=scales, colors=colors, line_style=line_style,
                         **kwargs)



import warnings
from pathlib import Path
import pytest

import numpy as np
from astroquery.mast import Observations
from stdatamodels.jwst.datamodels.dqflags import pixel as pixel_jwst
from glue.core.subset import RectangularROI

from jdaviz.configs.imviz.plugins.parsers import HAS_ROMAN_DATAMODELS
from jdaviz.configs.default.plugins.data_quality.dq_utils import (
    load_flag_map, write_flag_map
)


@pytest.mark.parametrize(
    "mission_or_instrument, flag, name,",
    [['jwst', 26, 'OPEN'],
     ['roman', 26, 'RESERVED_5'],
     ['jwst', 15, 'TELEGRAPH'],
     ['roman', 15, 'TELEGRAPH']]
)
def test_flag_map_utils(mission_or_instrument, flag, name):
    flag_map = load_flag_map(mission_or_instrument)
    assert flag_map[flag]['name'] == name


def test_load_write_load(tmp_path):
    # load the JWST flag map
    flag_map = load_flag_map('jwst')

    # write out the flag map to a temporary CSV file
    path = tmp_path / 'test_flag_map.csv'
    write_flag_map(flag_map, path)

    # load that temporary CSV file back in
    reloaded_flag_map = load_flag_map(path=path)

    # confirm all values round-trip successfully:
    for flag, orig_value in flag_map.items():
        assert orig_value == reloaded_flag_map[flag]


def test_jwst_against_stdatamodels():
    # compare our flag map against the flag map dictionary in `stdatamodels`:
    flag_map_loaded = load_flag_map('jwst')

    flag_map_expected = {
        int(np.log2(flag)): {'name': name}
        for name, flag in pixel_jwst.items() if flag > 0
    }

    # check no keys are missing in either flag map:
    assert len(set(flag_map_loaded.keys()) - set(flag_map_expected.keys())) == 0

    for flag in flag_map_expected:
        assert flag_map_loaded[flag]['name'] == flag_map_expected[flag]['name']


@pytest.mark.skipif(not HAS_ROMAN_DATAMODELS, reason="roman_datamodels is not installed")
def test_roman_against_rdm():
    # compare our flag map against the flag map dictionary in `roman_datamodels`:
    from roman_datamodels.dqflags import pixel as pixel_roman

    flag_map_loaded = load_flag_map('roman')

    flag_map_expected = {
        int(np.log2(p.value)): {'name': p.name}
        for p in pixel_roman if p.value > 0
    }

    # check no keys are missing in either flag map:
    assert len(set(flag_map_loaded.keys()) - set(flag_map_expected.keys())) == 0

    for flag in flag_map_expected:
        assert flag_map_loaded[flag]['name'] == flag_map_expected[flag]['name']


@pytest.mark.remote_data
def test_data_quality_plugin(imviz_helper, tmp_path):
    uri = "mast:JWST/product/jw01895001004_07101_00001_nrca3_cal.fits"
    local_path = str(tmp_path / Path(uri).name)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        imviz_helper.load_data(
            uri, cache=True, local_path=local_path, ext=('SCI', 'DQ')
        )

    assert len(imviz_helper.app.data_collection) == 2

    # this assumption is made in the DQ plugin (for now)
    assert imviz_helper.app.data_collection[-1].label.endswith('[DQ]')

    dq_plugin = imviz_helper.plugins['Data Quality']._obj

    # sci+dq layers are correctly identified
    expected_science_data, expected_dq_data = imviz_helper.app.data_collection
    assert dq_plugin.science_layer_selected == expected_science_data.label
    assert dq_plugin.dq_layer_selected == expected_dq_data.label

    # JWST data product identified as such:
    assert dq_plugin.flag_map_selected == 'JWST'

    # check flag 0 is a bad pixel in the JWST flag map:
    flag_map_selected = dq_plugin.flag_map_definitions_selected
    assert flag_map_selected[0]['name'] == 'DO_NOT_USE'
    assert flag_map_selected[0]['description'] == 'Bad pixel. Do not use.'

    # check default dq opacity is a fraction of sci data:
    sci_alpha = imviz_helper.default_viewer._obj.layers[0].state.alpha
    dq_alpha = imviz_helper.default_viewer._obj.layers[1].state.alpha
    assert dq_alpha == sci_alpha * dq_plugin.dq_layer_opacity

    plot_opts = imviz_helper.plugins['Plot Options']._obj

    # only the sci data appears in Plot Options:
    assert len(plot_opts.layer_items) == 1

    # check changes to sci opacity affect dq opacity
    new_sci_opacity = 0.5
    plot_opts.image_opacity_value = new_sci_opacity
    dq_alpha = imviz_helper.default_viewer._obj.layers[1].state.alpha
    assert dq_alpha == new_sci_opacity * dq_plugin.dq_layer_opacity

    # check that mouseover shows dq values on bad pixels (flag == 0):
    # check that mouseover shows dq values on bad pixels (flag == 0):
    viewer = imviz_helper.default_viewer._obj
    label_mouseover = imviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove', 'domain': {'x': 1366, 'y': 708}})

    # DQ features are labeled in the first line:
    label_mouseover_text = label_mouseover.as_text()[0]

    # bad pixels with flag == 0 have flux == NaN
    expected_flux_label = '+nan MJy / sr'
    assert expected_flux_label in label_mouseover_text

    # check that the decomposed DQ flag is at the end of the flux label's line:
    flux_label_idx = label_mouseover_text.index(expected_flux_label)
    associated_dq_layer = dq_plugin.get_dq_layers(viewer)[0]
    dq_data = associated_dq_layer.layer.get_data(associated_dq_layer.state.attribute)
    dq_val = dq_data[708, 1366]
    assert label_mouseover_text[flux_label_idx + len(expected_flux_label) + 1:] == f'(DQ: {dq_val:.0f})'  # noqa: E501

    # check that a flagged pixel that is not marked with the bit 0 has a flux in mouseover label:
    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove', 'domain': {'x': 1371, 'y': 715}})
    label_mouseover_text = label_mouseover.as_text()[0]
    dq_val = dq_data[715, 1371]
    assert label_mouseover_text.split('+')[1].endswith(f'(DQ: {dq_val:.0f})')

    # check that a pixel without a DQ flag has no DQ mouseover label:
    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove', 'domain': {'x': 1361, 'y': 684}})
    label_mouseover_text = label_mouseover.as_text()[0]
    assert 'DQ' not in label_mouseover_text

    # set a bit filter, then clear it:
    assert len(dq_plugin.flags_filter) == 0
    dq_plugin.flags_filter = [0, 1]
    dq_plugin.vue_clear_flags_filter({})
    assert len(dq_plugin.flags_filter) == 0

    # hide all:
    dq_plugin.vue_hide_all_flags({})
    assert not any([flag['show'] for flag in dq_plugin.decoded_flags])

    # now show all:
    dq_plugin.vue_show_all_flags({})
    assert all([flag['show'] for flag in dq_plugin.decoded_flags])


@pytest.mark.remote_data
def test_data_quality_plugin_hst_wfc3(imviz_helper, tmp_path):

    # load HST/WFC3-UVIS observations:
    uri = "mast:HST/product/hst_17183_02_wfc3_uvis_g280_iexr02mt_flt.fits"
    download_path = str(tmp_path / Path(uri).name)
    Observations.download_file(uri, local_path=download_path)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        imviz_helper.load_data(download_path, ext=('SCI', 'DQ'))

    assert len(imviz_helper.app.data_collection) == 2

    dq_plugin = imviz_helper.plugins['Data Quality']._obj

    # mission+instrument+detector identified:
    assert dq_plugin.flag_map_selected == 'HST/WFC3-UVIS'

    flag_map_selected = dq_plugin.flag_map_definitions_selected
    assert flag_map_selected[0]['description'] == 'Reed Solomon decoding error'


@pytest.mark.remote_data
def test_data_quality_plugin_hst_acs(imviz_helper, tmp_path):
    # load HST/ACS observations:
    uri = "mast:HST/product/hst_16968_01_acs_wfc_f606w_jezz01l3_flt.fits"
    download_path = str(tmp_path / Path(uri).name)
    Observations.download_file(uri, local_path=download_path)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        imviz_helper.load_data(download_path, ext=('SCI', 'DQ'))

    assert len(imviz_helper.app.data_collection) == 2

    dq_plugin = imviz_helper.plugins['Data Quality']._obj

    # mission+instrument identified:
    assert dq_plugin.flag_map_selected == 'HST/ACS'

    flag_map_selected = dq_plugin.flag_map_definitions_selected
    assert (
        flag_map_selected[0]['description'] ==
        'Reed-Solomon decoding error; e.g. data lost during compression.'
    )


@pytest.mark.remote_data
def test_cubeviz_layer_visibility_bug(cubeviz_helper, tmp_path):
    # regression test for bug:
    uri = "mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_s3d.fits"
    download_path = str(tmp_path / Path(uri).name)
    Observations.download_file(uri, local_path=download_path)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        cubeviz_helper.load_data(download_path)

    # create a moment map:
    mm = cubeviz_helper.plugins['Moment Maps']
    mm.n_moment = 1
    mm.reference_wavelength = 6
    mm._obj.add_to_viewer_selected = 'uncert-viewer'
    mm.calculate_moment()

    # add the moment map to the flux viewer
    dc = cubeviz_helper.app.data_collection
    viewer = cubeviz_helper.app.get_viewer('flux-viewer')
    viewer.add_data(dc[-1])

    # create a spatial subset in the flux-viewer
    roi = RectangularROI(22, 27, 22, 30)
    cubeviz_helper.plugins['Subset Tools'].import_region(roi)

    # toggle layer visibility, this used to trigger an AttributeError:
    cubeviz_helper.app.set_data_visibility('flux-viewer', dc[-1].label)
    cubeviz_helper.app.set_data_visibility('flux-viewer', dc[0].label)



# Tests data loading in the Mosviz Jdaviz configuration

from zipfile import ZipFile

import numpy as np
import pytest
from astropy.nddata import CCDData
from specutils import Spectrum1D

from jdaviz.utils import PRIHDR_KEY


def test_load_spectrum1d(mosviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    mosviz_helper.load_data(spectra_1d=spectrum1d, spectra_1d_label=label)

    assert len(mosviz_helper.app.data_collection) == 2
    assert mosviz_helper.app.data_collection[0].label == "MOS Table"
    dc_1 = mosviz_helper.app.data_collection[1]
    assert dc_1.label == label
    assert dc_1.meta['uncertainty_type'] == 'std'

    table = mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name)
    table.widget_table.vue_on_row_clicked(0)

    data = mosviz_helper.app.get_viewer(mosviz_helper._default_spectrum_viewer_reference_name
                                        ).data()

    assert len(data) == 1
    assert isinstance(data[0], Spectrum1D)

    with pytest.raises(AttributeError):
        mosviz_helper.load_1d_spectra([1, 2, 3])


def test_load_image(mosviz_helper, mos_image):
    label = "Test Image"
    mosviz_helper.load_images(mos_image, data_labels=label, add_redshift_column=True)

    assert len(mosviz_helper.app.data_collection) == 2
    assert mosviz_helper.app.data_collection[0].label == "MOS Table"
    dc_1 = mosviz_helper.app.data_collection[1]
    assert dc_1.label == f"{label} 0"
    assert PRIHDR_KEY not in dc_1.meta
    assert dc_1.meta['RADESYS'] == 'ICRS'

    table = mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name)
    table.widget_table.vue_on_row_clicked(0)

    data = mosviz_helper.app.get_viewer(mosviz_helper._default_image_viewer_reference_name
                                        ).data(cls=CCDData)

    assert len(data) == 1
    dataval = data[0]
    assert isinstance(dataval, CCDData)
    assert dataval.shape == (55, 55)


def test_load_spectrum_collection(mosviz_helper, spectrum_collection):
    labels = [f"Test Spectrum Collection {i}" for i in range(5)]
    mosviz_helper.load_1d_spectra(spectrum_collection, data_labels=labels, add_redshift_column=True)

    # +1 for the table viewer
    assert len(mosviz_helper.app.data_collection) == len(spectrum_collection) + 1
    assert mosviz_helper.app.data_collection[0].label == "MOS Table"
    dc_1 = mosviz_helper.app.data_collection[1]
    assert dc_1.label == labels[0]
    assert dc_1.meta['uncertainty_type'] == 'std'

    table = mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name)
    table.select_row(0)

    data = mosviz_helper.app.get_viewer(mosviz_helper._default_spectrum_viewer_reference_name
                                        ).data()

    assert len(data) == 1
    assert isinstance(data[0], Spectrum1D)


def test_load_list_of_spectrum1d(mosviz_helper, spectrum1d):
    spectra = [spectrum1d] * 3

    labels = [f"Test Spectrum 1D {i}" for i in range(3)]
    mosviz_helper.load_1d_spectra(spectra, data_labels=labels, add_redshift_column=True)

    assert len(mosviz_helper.app.data_collection) == 4
    assert mosviz_helper.app.data_collection[0].label == "MOS Table"
    dc_1 = mosviz_helper.app.data_collection[1]
    assert dc_1.label == labels[0]
    assert dc_1.meta['uncertainty_type'] == 'std'

    table = mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name)
    table.widget_table.vue_on_row_clicked(0)

    data = mosviz_helper.app.get_viewer(mosviz_helper._default_spectrum_viewer_reference_name
                                        ).data()

    assert len(data) == 1
    assert isinstance(data[0], Spectrum1D)


def test_load_mos_spectrum2d(mosviz_helper, mos_spectrum2d):

    label = "Test 2D Spectrum"
    mosviz_helper.load_data(spectra_2d=mos_spectrum2d, spectra_2d_label=label)

    assert len(mosviz_helper.app.data_collection) == 2
    assert mosviz_helper.app.data_collection[0].label == "MOS Table"
    dc_1 = mosviz_helper.app.data_collection[1]
    assert dc_1.label == label
    assert dc_1.meta['INSTRUME'] == 'nirspec'

    table = mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name)
    table.widget_table.vue_on_row_clicked(0)

    data = mosviz_helper.app.get_viewer(mosviz_helper._default_spectrum_2d_viewer_reference_name
                                        ).data()

    assert len(data) == 1
    assert data[0].shape == (1024, 15)


@pytest.mark.parametrize('label', [None, "Test Label"])
def test_load_multi_image_spec(mosviz_helper, mos_image, spectrum1d, mos_spectrum2d, label):
    spectra1d = [spectrum1d] * 3
    spectra2d = [mos_spectrum2d] * 3
    images = [mos_image] * 3

    mosviz_helper.load_data(spectra1d, spectra2d, images=images, images_label=label)

    assert mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name
                                        ).figure_widget.highlighted == 0
    assert len(mosviz_helper.app.data_collection) == 10

    qtable = mosviz_helper.to_table()
    if label is None:
        assert qtable["Images"][0] == "Image 0"
    else:
        assert qtable["Images"][0] == "Test Label 0"


def test_load_multi_image_and_spec1d_only(mosviz_helper, mos_image, spectrum1d):
    spectra1d = [spectrum1d] * 3
    images = [mos_image] * 3

    mosviz_helper.load_data(spectra_1d=spectra1d, images=images)

    assert mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name
                                        ).figure_widget.highlighted == 0
    assert len(mosviz_helper.app.data_collection) == 7


def test_load_multi_image_and_spec2d_only(mosviz_helper, mos_image, mos_spectrum2d):
    spectra2d = [mos_spectrum2d] * 3
    images = [mos_image] * 3

    mosviz_helper.load_data(spectra_2d=spectra2d, images=images)

    assert mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name
                                        ).figure_widget.highlighted == 0
    assert len(mosviz_helper.app.data_collection) == 7


@pytest.mark.parametrize('label', [None, "Test Label"])
def test_load_single_image_multi_spec(mosviz_helper, mos_image, spectrum1d, mos_spectrum2d, label):
    spectra1d = [spectrum1d] * 3
    spectra2d = [mos_spectrum2d] * 3

    image_viewer = mosviz_helper.app.get_viewer('image-viewer')
    spec1d_viewer = mosviz_helper.app.get_viewer('spectrum-viewer')
    spec2d_viewer = mosviz_helper.app.get_viewer('spectrum-2d-viewer')

    # Coordinates info panel should not crash even when nothing is loaded.
    label_mouseover = mosviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(image_viewer, {'event': 'mousemove',
                                                       'domain': {'x': 0, 'y': 0}})
    assert label_mouseover.as_text() == ('', '', '')

    # Test that loading is still possible after previous crash:
    # https://github.com/spacetelescope/jdaviz/issues/364
    with pytest.raises(ValueError, match="incompatible with the dimensions of this data:"):
        mosviz_helper.load_data(spectra1d, spectra2d, images=[mos_image, mos_image])

    mosviz_helper.load_data(spectra1d, spectra2d, images=mos_image, images_label=label)

    assert mosviz_helper.app.get_viewer(mosviz_helper._default_table_viewer_reference_name
                                        ).figure_widget.highlighted == 0
    assert len(mosviz_helper.app.data_collection) == 8

    qtable = mosviz_helper.to_table()
    if label is None:
        assert np.all(qtable["Images"] == "Shared Image")
    else:
        assert np.all(qtable["Images"] == "Test Label")
    assert len(qtable) == 3

    # Also check coordinates info panels for Mosviz image viewer.
    # 1D spectrum viewer panel is also tested in Specviz.
    # 2D spectrum viewer panel is also tested in Specviz2d.

    label_mouseover._viewer_mouse_event(image_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 0, 'y': 0}})
    assert label_mouseover.as_text() == ('Pixel x=000.0 y=000.0 Value +3.74540e-01 Jy',
                                         'World 00h20m07.1483s +04d59m30.8371s (ICRS)',
                                         '5.0297844783 4.9918991917 (deg)')
    assert label_mouseover.icon == 'a'

    label_mouseover._viewer_mouse_event(image_viewer,
                                        {'event': 'mousemove', 'domain': {'x': None, 'y': 0}})
    assert label_mouseover.as_text() == ('', '', '')
    assert label_mouseover.icon == ''

    label_mouseover._viewer_mouse_event(image_viewer,
                                        {'event': 'mousemove', 'domain': {'x': -1, 'y': 0}})
    assert label_mouseover.as_text() == ('Pixel x=-01.0 y=000.0',
                                         'World 00h20m07.1519s +04d59m30.8371s (ICRS)',
                                         '5.0297997183 4.9918991914 (deg)')

    label_mouseover._viewer_mouse_event(image_viewer, {'event': 'mouseleave'})
    assert label_mouseover.as_text() == ('', '', '')
    assert label_mouseover.icon == ''

    label_mouseover._viewer_mouse_event(spec2d_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 10, 'y': 100}})

    # Note: spectra2d Wave loaded in meters, but we respect one spectral unit, so the meters in
    # converted to Angstrom (the spectra1d spectral unit).
    assert label_mouseover.as_text() == ('Pixel x=00010.0 y=00100.0 Value +8.12986e-01 Jy',
                                         'Wave 1.10000e+05 Angstrom', '')
    assert label_mouseover.icon == 'c'

    # need to trigger a mouseleave or mouseover to reset the traitlets
    label_mouseover._viewer_mouse_event(spec1d_viewer, {'event': 'mouseenter'})
    label_mouseover._viewer_mouse_event(spec1d_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 7000, 'y': 170}})
    assert label_mouseover.as_text() == ('Cursor 7.00000e+03, 1.70000e+02',
                                         'Wave 6.88889e+03 Angstrom (4 pix)',
                                         'Flux 1.35436e+01 Jy')
    assert label_mouseover.icon == 'b'


def test_zip_error(mosviz_helper, tmp_path):
    '''
    Zipfiles are explicitly and intentionally not supported. This test confirms a TypeError is
    raised if the user tries to supply a Zipfile and expects Mosviz to autoextract.
    '''
    zip_path = tmp_path / "jdaviz_test_zip.zip"
    zip = ZipFile(zip_path, mode='w')
    zip.close()

    with pytest.raises(TypeError, match="Please extract"):
        mosviz_helper.load_data(directory=str(zip_path))


def test_invalid_inputs(mosviz_helper):
    with pytest.raises(NotImplementedError, match=r".*not a directory"):
        mosviz_helper.load_data(directory="foo")

    with pytest.raises(NotImplementedError, match="Please set valid values"):
        mosviz_helper.load_data()



from contextlib import nullcontext
from functools import cached_property

from astropy import units as u
from glue.core.subset_group import GroupedSubset
from glue_jupyter.bqplot.image import BqplotImageView
from specutils import Spectrum1D
from traitlets import List, Unicode, observe, Bool

from jdaviz.configs.default.plugins.viewers import JdavizProfileView
from jdaviz.core.custom_units_and_equivs import _eqv_flux_to_sb_pixel, _eqv_pixar_sr
from jdaviz.core.events import GlobalDisplayUnitChanged, AddDataMessage, SliceValueUpdatedMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, UnitSelectPluginComponent,
                                        SelectPluginComponent, PluginUserApi)
from jdaviz.core.unit_conversion_utils import (create_equivalent_spectral_axis_units_list,
                                               create_equivalent_flux_units_list,
                                               check_if_unit_is_per_solid_angle,
                                               create_equivalent_angle_units_list,
                                               supported_sq_angle_units)

__all__ = ['UnitConversion']


def _valid_glue_display_unit(unit_str, sv, axis='x'):
    # need to make sure the unit string is formatted according to the list of valid choices
    # that glue will accept (may not be the same as the defaults of the installed version of
    # astropy)
    if not unit_str:
        return unit_str
    unit_u = u.Unit(unit_str)
    choices_str = getattr(sv.state.__class__, f'{axis}_display_unit').get_choices(sv.state)
    choices_str = [choice for choice in choices_str if choice is not None]
    choices_u = [u.Unit(choice) for choice in choices_str]
    if unit_u not in choices_u:
        raise ValueError(f"{unit_str} could not find match in valid {axis} display units {choices_str}")  # noqa
    ind = choices_u.index(unit_u)
    return choices_str[ind]


def _flux_to_sb_unit(flux_unit, angle_unit):
    if angle_unit not in supported_sq_angle_units(as_strings=True):
        sb_unit = flux_unit
    else:
        # str > unit > str to remove formatting inconsistencies with
        # parentheses/order of units/etc
        sb_unit = (u.Unit(flux_unit) / u.Unit(angle_unit)).to_string()

    return sb_unit


@tray_registry('g-unit-conversion', label="Unit Conversion",
               viewer_requirements='spectrum')
class UnitConversion(PluginTemplateMixin):
    """
    The Unit Conversion plugin handles global app-wide unit-conversion.
    See the :ref:`Unit Conversion Plugin Documentation <unit-conversion>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``spectral_unit`` (:class:`~jdaviz.core.template_mixin.UnitSelectPluginComponent`):
      Global unit to use for all spectral axes.
    * ``flux_unit`` (:class:`~jdaviz.core.template_mixin.UnitSelectPluginComponent`):
      Global display unit for flux axis.
    * ``angle_unit`` (:class:`~jdaviz.core.template_mixin.UnitSelectPluginComponent`):
      Solid angle unit.
    * ``sb_unit`` (str): Read-only property for the current surface brightness unit,
      derived from the set values of ``flux_unit`` and ``angle_unit``.
    * ``spectral_y_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Select the y-axis physical type for the spectrum-viewer (applicable only to Cubeviz).
    * ``spectral_y_unit``: Read-only property for the current y-axis unit in the spectrum-viewer,
      either ``flux_unit`` or ``sb_unit`` depending on the selected ``spectral_y_type``
      (applicable only to Cubeviz).
    """
    template_file = __file__, "unit_conversion.vue"

    has_spectral = Bool(False).tag(sync=True)
    spectral_unit_items = List().tag(sync=True)
    spectral_unit_selected = Unicode().tag(sync=True)

    has_flux = Bool(False).tag(sync=True)
    flux_unit_items = List().tag(sync=True)
    flux_unit_selected = Unicode().tag(sync=True)

    has_angle = Bool(False).tag(sync=True)
    angle_unit_items = List().tag(sync=True)
    angle_unit_selected = Unicode().tag(sync=True)

    has_sb = Bool(False).tag(sync=True)
    sb_unit_selected = Unicode().tag(sync=True)

    has_time = Bool(False).tag(sync=True)
    time_unit_items = List().tag(sync=True)
    time_unit_selected = Unicode().tag(sync=True)

    spectral_y_type_items = List().tag(sync=True)
    spectral_y_type_selected = Unicode().tag(sync=True)

    # This shows an in-line warning message if False. This can be changed from
    # bool to unicode when we eventually handle inputing this value if it
    # doesn't exist in the FITS header
    pixar_sr_exists = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Convert the units of displayed physical quantities.'

        self._cached_properties = ['image_layers']

        if self.config not in ['specviz', 'specviz2d', 'cubeviz']:
            # TODO [mosviz] x_display_unit is not implemented in glue for image viewer
            # TODO [mosviz]: add to yaml file
            # TODO [cubeviz, slice]: slice indicator broken after changing spectral_unit
            # TODO: support for multiple viewers and handling of mixed state from glue (or does
            # this force all to sync?)
            self.disabled_msg = f'This plugin is temporarily disabled in {self.config}. Effort to improve it is being tracked at GitHub Issue 1972.'  # noqa

        self.session.hub.subscribe(self, AddDataMessage,
                                   handler=self._on_add_data_to_viewer)
        self.session.hub.subscribe(self, SliceValueUpdatedMessage,
                                   handler=self._on_slice_changed)

        self.has_spectral = self.config in ('specviz', 'cubeviz', 'specviz2d', 'mosviz')
        self.spectral_unit = UnitSelectPluginComponent(self,
                                                       items='spectral_unit_items',
                                                       selected='spectral_unit_selected')
        self.spectral_unit.choices = create_equivalent_spectral_axis_units_list(u.Hz)

        self.has_flux = self.config in ('specviz', 'cubeviz', 'specviz2d', 'mosviz')
        self.flux_unit = UnitSelectPluginComponent(self,
                                                   items='flux_unit_items',
                                                   selected='flux_unit_selected')
        # NOTE: will switch to count only if first data loaded into viewer in in counts
        # initialize flux choices to empty list, will be populated when data is loaded
        self.flux_unit.choices = []

        self.has_angle = self.config in ('cubeviz', 'specviz', 'mosviz', 'specviz2d')
        self.angle_unit = UnitSelectPluginComponent(self,
                                                    items='angle_unit_items',
                                                    selected='angle_unit_selected')
        # NOTE: will switch to pix2 only if first data loaded into viewer is in pix2 units
        # initialize angle unit choices to empty list, will be populated when data is loaded
        self.angle_unit.choices = []

        self.has_sb = self.has_angle or self.config in ('imviz',)
        # NOTE: sb_unit is read_only, exposed through sb_unit property

        self.has_time = False
        self.time_unit = UnitSelectPluginComponent(self,
                                                   items='time_unit_items',
                                                   selected='time_unit_selected')

        self.spectral_y_type = SelectPluginComponent(self,
                                                     items='spectral_y_type_items',
                                                     selected='spectral_y_type_selected')

    @property
    def user_api(self):
        expose = []
        readonly = []
        if self.has_spectral:
            expose += ['spectral_unit']
        if self.has_flux:
            expose += ['flux_unit']
        if self.has_angle:
            expose += ['angle_unit']
        if self.has_sb:
            readonly = ['sb_unit']
        if self.has_time:
            expose += ['time_unit']
        if self.config == 'cubeviz':
            expose += ['spectral_y_type', 'spectral_y_unit']
        return PluginUserApi(self, expose=expose, readonly=readonly)

    @property
    def sb_unit(self):
        # expose selected surface-brightness unit as read-only
        # (rather than exposing a select object)
        return self.sb_unit_selected

    @property
    def spectral_y_unit(self):
        return self.sb_unit_selected if self.spectral_y_type_selected == 'Surface Brightness' else self.flux_unit_selected  # noqa

    @cached_property
    def image_layers(self):
        return [layer
                for viewer in self._app._viewer_store.values() if isinstance(viewer, BqplotImageView)  # noqa
                for layer in viewer.layers]

    def _on_add_data_to_viewer(self, msg):
        # toggle warning message for cubes without PIXAR_SR defined
        if self.config == 'cubeviz':
            # NOTE: this assumes data_collection[0] is the science (flux/sb) cube
            if (
                len(self.app.data_collection) > 0
                and not self.app.data_collection[0].meta.get('PIXAR_SR')
            ):
                self.pixar_sr_exists = False

        viewer = msg.viewer

        if (not len(self.spectral_unit_selected)
                or not len(self.flux_unit_selected)
                or not len(self.angle_unit_selected)
                or (self.config == 'cubeviz' and not len(self.spectral_y_type_selected))):
            data_obj = msg.data.get_object()
            if isinstance(data_obj, Spectrum1D):

                self.spectral_unit._addl_unit_strings = self.spectrum_viewer.state.__class__.x_display_unit.get_choices(self.spectrum_viewer.state)  # noqa
                if not len(self.spectral_unit_selected):
                    try:
                        self.spectral_unit.selected = str(data_obj.spectral_axis.unit)
                    except ValueError:
                        self.spectral_unit.selected = ''

                angle_unit = check_if_unit_is_per_solid_angle(data_obj.flux.unit, return_unit=True)
                flux_unit = data_obj.flux.unit if angle_unit is None else data_obj.flux.unit * angle_unit  # noqa

                if not self.flux_unit_selected:
                    self.flux_unit.choices = create_equivalent_flux_units_list(flux_unit)
                    try:
                        self.flux_unit.selected = str(flux_unit)
                    except ValueError:
                        self.flux_unit.selected = ''

                if not self.angle_unit_selected:
                    self.angle_unit.choices = create_equivalent_angle_units_list(angle_unit)

                    try:
                        if angle_unit is None:
                            if self.config in ['specviz', 'specviz2d']:
                                self.has_angle = False
                                self.has_sb = False
                            else:
                                # default to pix2 if input data is not in surface brightness units
                                # TODO: for cubeviz, should we check the cube itself?
                                self.angle_unit.selected = 'pix2'
                        else:
                            self.angle_unit.selected = str(angle_unit)
                    except ValueError:
                        self.angle_unit.selected = ''

                if (not len(self.spectral_y_type_selected)
                        and isinstance(viewer, JdavizProfileView)):
                    # set spectral_y_type_selected to 'Flux'
                    # if the y-axis unit is not per solid angle
                    self.spectral_y_type.choices = ['Surface Brightness', 'Flux']
                    if angle_unit is None:
                        self.spectral_y_type_selected = 'Flux'
                    else:
                        self.spectral_y_type_selected = 'Surface Brightness'

                # setting default values will trigger the observes to set the units
                # in _on_unit_selected, so return here to avoid setting twice
                return

        # TODO: when enabling unit-conversion in rampviz, this may need to be more specific
        # or handle other cases for ramp profile viewers
        if isinstance(viewer, JdavizProfileView):
            if (viewer.state.x_display_unit == self.spectral_unit_selected
                    and viewer.state.y_display_unit == self.spectral_y_unit):
                # data already existed in this viewer and display units were already set
                return

            # this spectral viewer was empty (did not have display units set yet),˜
            # but global selections are available in the plugin,
            # so we'll set them to the viewer here
            viewer.state.x_display_unit = self.spectral_unit_selected
            # _handle_spectral_y_unit will call viewer.set_plot_axes()
            self._handle_spectral_y_unit()

        elif isinstance(viewer, BqplotImageView):
            # set the attribute display unit (contour and stretch units) for the new layer
            # NOTE: this assumes that all image data is coerced to surface brightness units
            layers = [lyr for lyr in msg.viewer.layers if lyr.layer.data.label == msg.data.label]
            self._handle_attribute_display_unit(self.sb_unit_selected, layers=layers)
            self._clear_cache('image_layers')

    def _on_slice_changed(self, msg):
        if self.config != "cubeviz":
            return
        self._cube_wave = u.Quantity(msg.value, msg.value_unit)

    @observe('spectral_unit_selected', 'flux_unit_selected',
             'angle_unit_selected', 'sb_unit_selected',
             'time_unit_selected')
    def _on_unit_selected(self, msg):
        """
        When any user selection is made, update the relevant viewer(s) with the new unit,
        and then emit a GlobalDisplayUnitChanged message to notify other plugins of the change.
        """
        if not len(msg.get('new', '')):
            # empty string, nothing to set yet
            return

        axis = msg.get('name').split('_')[0]

        if axis == 'spectral':
            xunit = _valid_glue_display_unit(self.spectral_unit.selected, self.spectrum_viewer, 'x')
            self.spectrum_viewer.state.x_display_unit = xunit
            self.spectrum_viewer.set_plot_axes()

        elif axis == 'flux':
            # handle spectral y-unit first since that is a more apparent change to the user
            # and feels laggy if it is done later
            if self.spectral_y_type_selected == 'Flux':
                self._handle_spectral_y_unit()

            if len(self.angle_unit_selected):
                # NOTE: setting sb_unit_selected will call this method again with axis=='sb',
                # which in turn will call _handle_attribute_display_unit,
                # _handle_spectral_y_unit (if spectral_y_type_selected == 'Surface Brightness'),
                #  and send a second GlobalDisplayUnitChanged message for sb
                self.sb_unit_selected = _flux_to_sb_unit(self.flux_unit.selected,
                                                         self.angle_unit.selected)

        elif axis == 'angle':
            if len(self.flux_unit_selected):
                # NOTE: setting sb_unit_selected will call this method again with axis=='sb',
                # which in turn will call _handle_attribute_display_unit,
                # _handle_spectral_y_unit (if spectral_y_type_selected == 'Surface Brightness'),
                #  and send a second GlobalDisplayUnitChanged message for sb
                self.sb_unit_selected = _flux_to_sb_unit(self.flux_unit.selected,
                                                         self.angle_unit.selected)

        elif axis == 'sb':
            # handle spectral y-unit first since that is a more apparent change to the user
            # and feels laggy if it is done later
            if self.spectral_y_type_selected == 'Surface Brightness':
                self._handle_spectral_y_unit()

            self._handle_attribute_display_unit(self.sb_unit_selected)

        # custom axes downstream can override _on_unit_selected if anything needs to be
        # processed before the GlobalDisplayUnitChanged message is broadcast

        # axis (first) argument will be one of: spectral, flux, angle, sb, time
        self.hub.broadcast(GlobalDisplayUnitChanged(axis,
                           msg.new, sender=self))

    @observe('spectral_y_type_selected')
    def _handle_spectral_y_unit(self, *args):
        """
        When the spectral_y_type is changed, or the unit corresponding to the
        currently selected spectral_y_type is changed, update the y-axis of
        the spectrum viewer with the new unit, and then emit a
        GlobalDisplayUnitChanged message to notify
        """
        yunit = _valid_glue_display_unit(self.spectral_y_unit, self.spectrum_viewer, 'y')
        if self.spectrum_viewer.state.y_display_unit == yunit:
            self.spectrum_viewer.set_plot_axes()
            return
        try:
            self.spectrum_viewer.state.y_display_unit = yunit
        except ValueError:
            # may not be data in the viewer, or unit may be incompatible
            pass
        else:
            self.spectrum_viewer.set_plot_axes()

            # broadcast that there has been a change in the spectrum viewer y axis,
            self.hub.broadcast(GlobalDisplayUnitChanged('spectral_y',
                                                        yunit,
                                                        sender=self))

    def _handle_attribute_display_unit(self, attr_unit, layers=None):
        """
        Update the per-layer attribute display unit in glue for image viewers
        (updating stretch and contour units).
        """
        if layers is None:
            layers = self.image_layers

        for layer in layers:
            # DQ layer doesn't play nicely with this attribute
            if "DQ" in layer.layer.label or isinstance(layer.layer, GroupedSubset):
                continue
            elif ("flux" not in [str(c) for c in layer.layer.components]
                    or u.Unit(layer.layer.get_component("flux").units).physical_type != 'surface brightness'):  # noqa
                continue
            if hasattr(layer.state, 'attribute_display_unit'):
                if self.config == "cubeviz":
                    ctx = u.set_enabled_equivalencies(
                        u.spectral() + u.spectral_density(self._cube_wave) +
                        _eqv_flux_to_sb_pixel() +
                        _eqv_pixar_sr(layer.layer.meta.get('_pixel_scale_factor', 1)))
                else:
                    ctx = nullcontext()
                with ctx:
                    layer.state.attribute_display_unit = _valid_glue_display_unit(
                        attr_unit, layer, 'attribute')



from functools import cached_property

from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry


@loader_parser_registry('object')
class ObjectParser(BaseParser):
    # pass through an object from the object resolver directly to the importers
    @property
    def is_valid(self):
        return self.input is not None

    @cached_property
    def output(self):
        return self.input



from astropy.utils.decorators import deprecated

from jdaviz.configs.default.plugins.line_lists.line_list_mixin import LineListMixin
from jdaviz.configs.specviz import Specviz
from jdaviz.core.events import AddDataMessage, SnackbarMessage
from jdaviz.core.helpers import CubeConfigHelper
from jdaviz.configs.cubeviz.plugins.viewers import CubevizImageView

__all__ = ['Cubeviz']


_spectral_axis_names = ["Wave", "Wavelength", "Freq", "Frequency",
                        "Wavenumber", "Velocity", "Energy"]


class Cubeviz(CubeConfigHelper, LineListMixin):
    """Cubeviz Helper class"""
    _default_configuration = 'cubeviz'
    _default_spectrum_viewer_reference_name = "spectrum-viewer"
    _default_uncert_viewer_reference_name = "uncert-viewer"
    _default_flux_viewer_reference_name = "flux-viewer"
    _default_image_viewer_reference_name = "image-viewer"
    _cube_viewer_default_label = _default_flux_viewer_reference_name

    _loaded_flux_cube = None
    _loaded_uncert_cube = None
    _loaded_mask_cube = None
    _cube_viewer_cls = CubevizImageView

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app.hub.subscribe(self, AddDataMessage,
                               handler=self._set_spectrum_x_axis)

    def _set_spectrum_x_axis(self, msg):
        viewer = self.app.get_viewer(self._default_spectrum_viewer_reference_name)
        if msg.viewer_id != viewer.reference_id:
            return
        ref_data = viewer.state.reference_data
        if ref_data and ref_data.ndim == 3:
            for att_name in _spectral_axis_names:
                if att_name in ref_data.component_ids():
                    if viewer.state.x_att != ref_data.id[att_name]:
                        viewer.state.x_att = ref_data.id[att_name]
                        viewer.state.reset_limits()
                    break
            else:
                viewer.state.x_att = ref_data.id["Pixel Axis 2 [x]"]
                viewer.state.reset_limits()

    def load_data(self, data, data_label=None, override_cube_limit=False, **kwargs):
        """
        Load and parse a data cube with Cubeviz.
        (Note that only one cube may be loaded per Cubeviz instance.)

        Parameters
        ----------
        data : str, `~astropy.io.fits.HDUList`, `~specutils.Spectrum1D`, or ndarray
            A string file path, astropy FITS object pointing to the
            data cube, a spectrum object, or a Numpy array cube.
            If plain array is given, axes order must be ``(x, y, z)``.
        data_label : str or `None`
            Data label to go with the given data. If not given,
            one will be automatically generated.
        override_cube_limit : bool
            Override internal cube count limitation and load the data anyway.
            Setting this to `True` is not recommended unless you know what
            you are doing.
        **kwargs : dict
            Extra keywords accepted by Jdaviz application-level parser.

        """
        if not override_cube_limit and len(self.app.state.data_items) != 0:
            raise RuntimeError('Only one cube may be loaded per Cubeviz instance')
        if data_label:
            kwargs['data_label'] = data_label

        super().load_data(data, parser_reference="cubeviz-data-parser", **kwargs)

        if 'Spectral Extraction' not in self.plugins:  # pragma: no cover
            msg = SnackbarMessage(
                "Automatic spectral extraction requires the Spectral Extraction plugin to be enabled",  # noqa
                color='error', sender=self, timeout=10000)
            self.app.hub.broadcast(msg)
        else:
            try:
                self.plugins['Spectral Extraction']._obj._extract_in_new_instance(auto_update=False, add_data=True)  # noqa
            except Exception:
                msg = SnackbarMessage(
                    "Automatic spectrum extraction for the entire cube failed."
                    " See the spectral extraction plugin to perform a custom extraction",
                    color='error', sender=self, timeout=10000)
            else:
                msg = SnackbarMessage(
                    "The extracted 1D spectrum was generated automatically for the entire cube."
                    " See the spectral extraction plugin for details or to"
                    " perform a custom extraction.",
                    color='warning', sender=self, timeout=10000)
            self.app.hub.broadcast(msg)

    @deprecated(since="4.2", alternative="plugins['Slice'].value")
    def select_wavelength(self, wavelength):
        """
        Select the slice closest to the provided wavelength.

        Parameters
        ----------
        wavelength : float
            Wavelength to select in units of the x-axis of the spectrum.  The nearest slice will
            be selected if "snap to slice" is enabled in the slice plugin.
        """
        if not isinstance(wavelength, (int, float)):
            raise TypeError("wavelength must be a float or int")
        self.select_slice(wavelength)

    @property
    def specviz(self):
        """
        A Specviz helper (:class:`~jdaviz.configs.specviz.helper.Specviz`) for the Jdaviz
        application that is wrapped by Cubeviz.
        """
        if not hasattr(self, '_specviz'):
            self._specviz = Specviz(app=self.app)
        return self._specviz

    def get_data(self, data_label=None, spatial_subset=None, spectral_subset=None,
                 cls=None, use_display_units=False):
        """
        Returns data with name equal to ``data_label`` of type ``cls`` with subsets applied from
        ``spectral_subset``, if applicable.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        spatial_subset : str, optional
            Spatial subset applied to data.  Only applicable if ``data_label`` points to a cube or
            image.  To extract a spectrum from a cube, use the spectral extraction plugin instead.
        spectral_subset : str, optional
            Spectral subset applied to data.
        cls : `~specutils.Spectrum1D`, `~astropy.nddata.CCDData`, optional
            The type that data will be returned as.
        use_display_units : bool, optional
            Specify whether the returned data is in native units or the current display units.

        Returns
        -------
        data : cls
            Data is returned as type cls with subsets applied.

        """
        return self._get_data(data_label=data_label, spatial_subset=spatial_subset,
                              spectral_subset=spectral_subset,
                              cls=cls, use_display_units=use_display_units)

    @deprecated(since="4.2", alternative="plugins['Aperture Photometry'].export_table()")
    def get_aperture_photometry_results(self):
        """Return aperture photometry results, if any.
        Results are calculated using :ref:`cubeviz-aper-phot` plugin.

        Returns
        -------
        results : `~astropy.table.QTable` or `None`
            Photometry results if available or `None` otherwise.

        """
        return self.plugins['Aperture Photometry'].export_table()



import numpy as np
from traitlets import Bool, Unicode, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import ViewerAddedMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, ViewerSelectMixin, Plot,
                                        skip_if_no_updates_since_last_active)
from jdaviz.utils import get_top_layer_index


__all__ = ['LineProfileXY']


@tray_registry('imviz-line-profile-xy', label="Image Profiles (XY)")
class LineProfileXY(PluginTemplateMixin, ViewerSelectMixin):
    template_file = __file__, "line_profile_xy.vue"
    uses_active_status = Bool(True).tag(sync=True)

    plot_available = Bool(False).tag(sync=True)
    selected_x = FloatHandleEmpty('').tag(sync=True)
    selected_y = FloatHandleEmpty('').tag(sync=True)

    plot_across_x_widget = Unicode().tag(sync=True)
    plot_across_y_widget = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Plot line profiles across X and Y.'

        self.plot_across_x = Plot(self, name='across_x')
        self.plot_across_y = Plot(self, name='across_y')
        for plot in (self.plot_across_x, self.plot_across_y):
            # override default styling
            plot.figure.fig_margin = {'top': 60, 'bottom': 60, 'left': 65, 'right': 15}
            plot.viewer.axis_x.num_ticks = 5
            plot.viewer.axis_y.tick_format = '0.2e'
            plot.viewer.axis_y.label_offset = '55px'

        self.plot_across_x_widget = 'IPY_MODEL_'+self.plot_across_x.model_id
        self.plot_across_y_widget = 'IPY_MODEL_'+self.plot_across_y.model_id

        self.docs_description = "Press 'l' to plot line profiles across X and Y\
                                 under cursor. You can also manually enter X, Y\
                                 and then click PLOT."

        self.hub.subscribe(self, ViewerAddedMessage, handler=self._on_viewer_added)

    def reset_results(self):
        self.plot_available = False
        self.plot_across_x.update_style('line', visible=False)
        self.plot_across_x.clear_all_marks()
        self.plot_across_y.clear_all_marks()

    def _create_viewer_callbacks(self, viewer):
        if not self.is_active:
            return

        callback = self._viewer_callback(viewer, self._on_viewer_key_event)
        viewer.add_event_callback(callback, events=['keydown'])

    def _on_viewer_added(self, msg):
        self._create_viewer_callbacks(self.app.get_viewer_by_id(msg.viewer_id))

    @observe('is_active')
    def _is_active_changed(self, msg):
        # subscribe/unsubscribe to keypress events across all viewers
        for viewer in self.app._viewer_store.values():
            if not hasattr(viewer, 'figure'):
                # table viewer, etc
                continue
            callback = self._viewer_callback(viewer, self._on_viewer_key_event)

            if self.is_active:
                viewer.add_event_callback(callback, events=['keydown'])
            else:
                viewer.remove_event_callback(callback)

        # pass along the msg object so that @skip_if_no_updates_since_last_active can be used
        # to avoid re-drawing if no changes since the last time is_active was set
        self.vue_draw_plot(msg)

    def _on_viewer_key_event(self, viewer, data):
        if data['key'] == 'l':
            image = viewer.active_image_layer.layer
            x = data['domain']['x']
            y = data['domain']['y']
            if x is None or y is None:  # Out of bounds
                return
            x, y, _, _ = viewer._get_real_xy(image, x, y)
            self.selected_x = x
            self.selected_y = y
            self.viewer_selected = viewer.reference_id
            # TODO: remove manual calls to vue_draw_plot and trigger
            # by changes to selected_x/selected_y as well as viewer_selected
            self.vue_draw_plot()

    @observe("viewer_selected")
    @skip_if_no_updates_since_last_active()  # called with msg passed along from _is_active_changed
    def vue_draw_plot(self, msg={}):
        """Draw line profile plots for given Data across given X and Y indices (0-indexed)."""
        if not self.selected_x or not self.selected_y:
            return

        viewer = self.viewer.selected_obj
        i = get_top_layer_index(viewer)
        data = viewer.state.layers[i].layer

        x = int(round(self.selected_x))
        y = int(round(self.selected_y))

        nx = data.shape[1]
        ny = data.shape[0]
        if x < 0 or y < 0 or x >= nx or y >= ny:
            self.reset_results()
            return

        comp = data.get_component(data.main_components[0])
        if comp.units:
            y_label = comp.units
        else:
            y_label = 'Value'

        xy_limits = viewer._get_zoom_limits(data)
        x_limits = xy_limits[:, 0]
        y_limits = xy_limits[:, 1]
        x_min = max(int(x_limits.min()), 0)
        x_max = min(int(x_limits.max()), nx)
        y_min = max(int(y_limits.min()), 0)
        y_max = min(int(y_limits.max()), ny)

        self.plot_across_x.figure.title = f'X={x}'
        self.plot_across_x._update_data('line', x=range(comp.data.shape[0]), y=comp.data[:, x],
                                        reset_lims=False)
        zoomed_data_x = comp.data[y_min:y_max, x]
        if zoomed_data_x.size > 0:
            self.plot_across_x.set_limits(x_min=y_min,
                                          x_max=y_max,
                                          y_min=np.nanmin(zoomed_data_x) * 0.95,
                                          y_max=np.nanmax(zoomed_data_x) * 1.05)
        self.plot_across_x.update_style('line', line_visible=True,
                                        markers_visible=False, color='gray', size=10)
        self.plot_across_x.viewer.axis_x.label = 'Y (pix)'
        self.plot_across_x.viewer.axis_y.label = y_label

        self.plot_across_y.figure.title = f'Y={y}'
        self.plot_across_y._update_data('line', x=range(comp.data.shape[1]), y=comp.data[y, :],
                                        reset_lims=False)
        zoomed_data_y = comp.data[y, x_min:x_max]
        if zoomed_data_y.size > 0:
            self.plot_across_y.set_limits(x_min=x_min,
                                          x_max=x_max,
                                          y_min=np.nanmin(zoomed_data_y) * 0.95,
                                          y_max=np.nanmax(zoomed_data_y) * 1.05)
        self.plot_across_y.update_style('line', line_visible=True,
                                        markers_visible=False, color='gray', size=10)
        self.plot_across_y.viewer.axis_x.label = 'X (pix)'
        self.plot_across_y.viewer.axis_y.label = y_label

        self.plot_available = True



import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import InverseVariance
from specutils import Spectrum1D

from jdaviz.core.custom_units_and_equivs import SPEC_PHOTON_FLUX_DENSITY_UNITS


# On failure, should not crash; essentially a no-op.
@pytest.mark.parametrize(
    ('new_spectral_axis', 'new_flux', 'expected_spectral_axis', 'expected_flux'),
    [("fail", "erg / (s cm2 Angstrom)", "Angstrom", "erg / (s cm2 Angstrom)"),
     ("None", "fail", "Angstrom", "Jy"),
     ("micron", "fail", "micron", "Jy")])
def test_value_error_exception(specviz_helper, spectrum1d, new_spectral_axis, new_flux,
                               expected_spectral_axis, expected_flux):
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")
    viewer = specviz_helper.app.get_viewer("spectrum-viewer")
    plg = specviz_helper.plugins["Unit Conversion"]

    try:
        plg.spectral_unit = new_spectral_axis
    except ValueError as e:
        if "reverting selection to" not in repr(e):
            raise
    try:
        plg._obj.flux_unit_selected = new_flux
    except ValueError as e:
        if "reverting selection to" not in repr(e):
            raise

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit(expected_spectral_axis)
    assert u.Unit(viewer.state.y_display_unit) == u.Unit(expected_flux)


def test_initialize_specviz_sb(specviz_helper, spectrum1d):
    spec_sb = Spectrum1D(spectrum1d.flux/u.sr, spectrum1d.spectral_axis)
    specviz_helper.load_data(spec_sb, data_label="Test 1D Spectrum")
    plg = specviz_helper.plugins["Unit Conversion"]
    assert plg._obj.flux_unit == "Jy"
    assert plg._obj.spectral_y_type == "Surface Brightness"
    assert plg._obj.angle_unit == "sr"


@pytest.mark.parametrize('uncert', (False, True))
def test_conv_wave_only(specviz_helper, spectrum1d, uncert):
    if uncert is False:
        spectrum1d.uncertainty = None
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    viewer = specviz_helper.app.get_viewer("spectrum-viewer")
    plg = specviz_helper.plugins["Unit Conversion"]
    new_spectral_axis = "micron"
    plg.spectral_unit = new_spectral_axis

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit(new_spectral_axis)
    assert u.Unit(viewer.state.y_display_unit) == u.Unit('Jy')


@pytest.mark.parametrize('uncert', (False, True))
def test_conv_flux_only(specviz_helper, spectrum1d, uncert):
    if uncert is False:
        spectrum1d.uncertainty = None
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    viewer = specviz_helper.app.get_viewer("spectrum-viewer")
    plg = specviz_helper.plugins["Unit Conversion"]
    new_flux = "erg / (s cm2 Angstrom)"
    plg._obj.flux_unit_selected = new_flux

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit('Angstrom')
    assert u.Unit(viewer.state.y_display_unit) == u.Unit(new_flux)


@pytest.mark.parametrize('uncert', (False, True))
def test_conv_wave_flux(specviz_helper, spectrum1d, uncert):
    if uncert is False:
        spectrum1d.uncertainty = None
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    viewer = specviz_helper.app.get_viewer("spectrum-viewer")
    plg = specviz_helper.plugins["Unit Conversion"]
    new_spectral_axis = "micron"
    new_flux = "erg / (s cm2 Angstrom)"
    plg.spectral_unit = new_spectral_axis
    plg._obj.flux_unit_selected = new_flux

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit(new_spectral_axis)
    assert u.Unit(viewer.state.y_display_unit) == u.Unit(new_flux)


def test_conv_no_data(specviz_helper, spectrum1d):
    """plugin unit selections won't have valid choices yet, preventing
    attempting to set display units."""
    # spectrum not load is in Flux units, sb_unit and flux_unit
    # should be enabled, spectral_y_type should not be
    plg = specviz_helper.plugins["Unit Conversion"]
    with pytest.raises(ValueError, match="could not find match in valid x display units"):
        plg.spectral_unit = "micron"
    assert len(specviz_helper.app.data_collection) == 0

    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    # make sure we don't expose translations in Specviz
    assert hasattr(plg, 'flux_unit')
    assert hasattr(plg, 'angle_unit')
    assert not hasattr(plg, 'spectral_y_type')


def test_non_stddev_uncertainty(specviz_helper):
    flux = np.ones(10) * u.Jy
    stddev = 0.1
    var = stddev ** 2
    inv_var = np.ones(len(flux)) / var
    wavelength = np.linspace(1, 5, len(flux)) * u.um
    spec = Spectrum1D(
        flux,
        uncertainty=InverseVariance(inv_var),
        spectral_axis=wavelength
    )

    specviz_helper.load_data(spec)

    po = specviz_helper.plugins['Plot Options']
    po.uncertainty_visible = True

    # check that the stddev uncertainties are drawn:
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')
    np.testing.assert_allclose(
        np.abs(viewer.figure.marks[-1].y - viewer.figure.marks[-1].y.mean(0)),
        stddev
    )


@pytest.mark.parametrize("flux_unit, expected_choices", [(u.count, ['ct']),
                                                         (u.Jy, SPEC_PHOTON_FLUX_DENSITY_UNITS),
                                                         (u.nJy, SPEC_PHOTON_FLUX_DENSITY_UNITS + ['nJy'])])  # noqa
def test_flux_unit_choices(specviz_helper, flux_unit, expected_choices):
    """
    Test that cubes loaded with various flux units have the expected default
    flux unit selection in the unit conversion plugin, and that the list of
    convertable flux units in the dropdown is correct.
    """

    spec = Spectrum1D([1, 2, 3] * flux_unit, [4, 5, 6] * u.um)
    specviz_helper.load_data(spec)

    uc_plg = specviz_helper.plugins['Unit Conversion']

    assert uc_plg.flux_unit.selected == flux_unit.to_string()
    assert uc_plg.flux_unit.choices == expected_choices


def test_mosviz_profile_view_mouseover(specviz2d_helper):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum1D(flux=data*u.MJy, spectral_axis=data[3]*u.um)

    specviz2d_helper.load_data(spectrum2d)
    viewer = specviz2d_helper.app.get_viewer("spectrum-viewer")
    plg = specviz2d_helper.plugins["Unit Conversion"]

    # make sure we don't expose angle, sb, nor spectral-y units when native
    # units are in flux
    assert hasattr(plg, 'flux_unit')
    assert not hasattr(plg, 'angle_unit')
    assert not hasattr(plg, 'sb_unit')
    assert not hasattr(plg, 'spectral_y_type')

    label_mouseover = specviz2d_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 5, 'y': 3}})

    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.00000e+00 MJy')

    plg._obj.flux_unit_selected = 'Jy'
    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.00000e+06 Jy')

    # test mouseover when spectral density equivalencies are required for conversion
    plg._obj.flux_unit_selected = 'erg / (Angstrom s cm2)'
    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.99585e-08 erg / (Angstrom s cm2)')



from PyInstaller.utils.hooks import collect_data_files, copy_metadata, collect_submodules

hiddenimports = collect_submodules("scipy")
datas = collect_data_files('scipy')
datas += copy_metadata('scipy')
